#!/usr/bin/env python
# encoding: utf-8
#
# Copyright 2017 Jet Digital Inc. All rights reserved.
#------------------------------------------------------------------------------------------------
# Due to SSL/TLS security bugs and missing features, contemporary version of Python is required.
# Code was written to run on much older versions but circumventing the restriction is not recommended.
import sys
def _VERSION_STR(v):
    return '.'.join(tuple(str(x) for x in v))
_PYTHON_VERSION = (sys.version_info.major, sys.version_info.minor, sys.version_info.micro)
_V27_MIN = (2, 7, 4) # released 6 April 2013
_V32_MIN = (3, 2, 4) # released 7 April 2013
_V33_MIN = (3, 3, 1) # released 7 April 2013
if _PYTHON_VERSION < _V27_MIN:
    raise ImportError('Python version %s+ required. Using version %s.' % (_VERSION_STR(_V27_MIN), _VERSION_STR(_PYTHON_VERSION)))
if _PYTHON_VERSION >= (3, 0, 0) and _PYTHON_VERSION < _V32_MIN:
    raise ImportError('Python version %s+ required. Using version %s.' % (_VERSION_STR(_V32_MIN), _VERSION_STR(_PYTHON_VERSION)))
if _PYTHON_VERSION >= (3, 3, 0) and _PYTHON_VERSION < _V33_MIN:
    raise ImportError('Python version %s+ required. Using version %s.' % (_VERSION_STR(_V33_MIN), _VERSION_STR(_PYTHON_VERSION)))
#------------------------------------------------------------------------------------------------

import ssl

from jetstream.constants import DEFAULT_SERVER_ADDRESS, DEFAULT_API_PORT, DEFAULT_TIMEOUT_SECONDS, NULL
from jetstream.apiinterface import APIInterface
from jetstream.logginginterface import LoggingInterface
from jetstream.socketinterface import SocketInterface
from jetstream.syncinterface import SyncInterface

class JetStream(object):
    """
    The ``JetStream`` class serves as an *API interface factory*. Use this to instantiate an interface
    that may be used to issue commands to a JetStream server.

    Note that this factory, on its own, does not connect to a JetStream server, nor issue commands.
    You use the returned interfaces to connect and issue commands.

    Example:
        .. code-block:: python

            jetstream = JetStream()

            # Instantiate an API interface.
            api = jetstream.api(serverAddress, serverPort)

            # Connect
            with api:
                # Issue commands.
                api.server.getServerInfo()
                api.send.createManifest(...)
    """


    def __init__(self, socket_interface=SocketInterface, logging_interface=LoggingInterface, sync_interface=SyncInterface):
        self.logging_interface = logging_interface
        self.socket_interface = socket_interface
        self.sync_interface = sync_interface

    def api(self, serverAddress=DEFAULT_SERVER_ADDRESS, serverPort=DEFAULT_API_PORT, timeoutSeconds=DEFAULT_TIMEOUT_SECONDS, certificatePath=NULL, sslCertificateRequirements=ssl.CERT_NONE):
        """Get an API interface for a JetStream server.

        This does not connect to the server. Use the returned interface to connect first, then
        issue commands.

        Note:
            Timeout functionality is currently *unimplemented*.

        Args:
            serverAddress (str): The address (host name) on which the JetStream server is running.
            serverPort (int): The control port on which the server is listening.
            timeoutSeconds (float): A timeout, in seconds, used to terminate blocking commands. If unassigned, blocking commands will never unblock.
            certificatePath (str): Path to an SSL certificate.
            sslCertificateRequirements (int): Level of strictness when checking SSL certificates. Can be one of: ``ssl.CERT_NONE`` (0), ``ssl.CERT_OPTIONAL`` (1), or ``ssl.CERT_REQUIRED`` (2).

        Returns:
            APIInterface: An API interface.
        """
        socket_interface = self.__get_socket_interface(serverAddress, serverPort, timeoutSeconds, certificatePath, sslCertificateRequirements)
        sync_interface = self.__get_sync_interface(timeoutSeconds)
        logging_interface = self.logging_interface()
        return APIInterface(socket_interface, sync_interface, logging_interface)

    def api_listen(self, listenAddress='', listenPort=DEFAULT_API_PORT, timeoutSeconds=DEFAULT_TIMEOUT_SECONDS, certificatePath=NULL, sslCertificateRequirements=ssl.CERT_NONE):
        """Get an API interface that listens for connections from JetStream servers.

        This does not listen for connections. Use the returned interface to start
        listening first, then issue commands

        See :func:`api` for details.
        """

        socket_interface = self.__get_socket_interface(listenAddress, listenPort, timeoutSeconds, certificatePath, sslCertificateRequirements)
        sync_interface = self.__get_sync_interface(timeoutSeconds)
        logging_interface = self.logging_interface()

        return APIInterface(socket_interface, sync_interface, logging_interface, listen=True)

    def __get_socket_interface(self, serverAddress, serverPort, timeoutSeconds, certificatePath, sslCertificateRequirements):
        socket_interface = self.socket_interface()	
        socket_interface.serverAddress = serverAddress
        socket_interface.serverPort = serverPort
        socket_interface.timeoutSeconds = timeoutSeconds
        socket_interface.certificatePath = certificatePath
        socket_interface.sslCertificateRequirements = sslCertificateRequirements

        return socket_interface

    def __get_sync_interface(self, timeoutSeconds):
        sync_interface = self.sync_interface()
        sync_interface.timeoutSeconds = timeoutSeconds
        return sync_interface

# [backcompat]
Jetstream = JetStream
