#!/usr/bin/env python
# encoding: utf-8
#
# Copyright 2018 Jet Digital Inc. All rights reserved.
#------------------------------------------------------------------------------------------------

import functools
import inspect
import sys
from jetstream.errors import ApiVersionError

def __check_api_compatibility(func_name, api_interface, min_version=None, max_version=None):
    if hasattr(api_interface, 'api_version') and (api_interface.api_version is not None):
        if (min_version is not None) and (api_interface.api_version < min_version):
            raise ApiVersionError(func_name, min_version, api_interface.api_version)
        if (max_version is not None) and (api_interface.api_version > max_version):
            raise ApiVersionError(func_name, max_version, api_interface.api_version)

def jetapi_extract_param(key, value):
    if key == 'objId':
        try:
            return value['transferId']
        except (AttributeError, TypeError, KeyError):
            pass
        try:
            return value['token']
        except (AttributeError, TypeError, KeyError):
            pass
        try:
            return value['manifestId']
        except (AttributeError, TypeError, KeyError):
            pass
        try:
            return value['destinationId']
        except (AttributeError, TypeError, KeyError):
            pass
        return value
    else:
        try:
            return value[key]
        except (AttributeError, TypeError, KeyError):
            return value

def jetapi(min_api, max_api=None, auth=False, superuser=False, **jet_attribs):
    # pylint: disable=protected-access
    def wrapped(function):
        if sys.version_info[0] > 2:
            jet_argspec = inspect.getfullargspec(function) # pylint: disable=no-member,deprecated-method
        else:
            jet_argspec = inspect.getargspec(function) # pylint: disable=no-member,deprecated-method

        @functools.wraps(function)
        def wrapped_fn(inst, *args, **kwargs):
            # Check api compatibility.
            if sys.version_info[0] > 2:
                function_name=function.__name__
            else:
                function_name=function.func_name
            __check_api_compatibility(function_name, inst._api_interface, min_api, max_api)

            # Extract any decorated parameters from dictionaries.
            newargs = list(args)
            for idx, arg in enumerate(jet_argspec.args[1:]): # skip 'self'
                if arg in jet_attribs:
                    if idx < len(args):
                        newargs[idx] = jetapi_extract_param(arg, args[idx])
                    elif arg in kwargs:
                        kwargs[arg] = jetapi_extract_param(arg, kwargs[arg])
            result = function(inst, *newargs, **kwargs)
            return result
        wrapped_fn.jetapi = jet_attribs
        wrapped_fn.jetapi['min_api'] = min_api
        wrapped_fn.jetapi['auth'] = auth
        wrapped_fn.jetapi['superuser'] = superuser
        wrapped_fn.jetapi['argspec'] = jet_argspec
        return wrapped_fn
    return wrapped
