#!/usr/bin/env python
# encoding: utf-8
#
# Copyright 2017 Jet Digital Inc. All rights reserved.
#------------------------------------------------------------------------------------------------

import ssl
from jetstream.constants import NULL, API_VERSION
from jetstream.errors import JetStreamClientException, NotConnectedError
from jetstream.recvinterface import RecvInterface
from jetstream.sendinterface import SendInterface
from jetstream.serverinterface import ServerInterface

class APIInterface(object):
    """
    This is the primary API interface for JetStream. The methods on the API interface are used to
    establish a connection to a JetStream server.

    An ``APIInterface`` should be constructed from the ``JetStream`` :doc:`API factory <jetstream.api.JetStream>`.

    Example:
        .. code-block:: python

            api_interface = JetStream().api()

    There are two ways to connect to a JetStream server.

    First, you can use the ``connect`` and ``disconnect`` methods.

    Example:
        .. code-block:: python

            api_interface.connect()

            # Issue commands.

            api_interface.disconnect()

    Or, you can use the interface as a *context manager*.

    Example:
        .. code-block:: python

            with api_interface:
                # Issue commands.

    The API interface exposes additional interfaces to control the different aspects of a JetStream
    server. These interfaces are accessed through their respective *namespaces*.

    Use ``api.send`` to access methods related to sending files.

    Example:
        .. code-block:: python

            api = JetStream().api()
            api.send.<method>()

    .. seealso::
        - :doc:`jetstream.sendinterface.SendInterface`

    Use ``api.recv`` to access methods related to receiving files.

    Example:
        .. code-block:: python

            api = JetStream().api()
            api.recv.<method>()

    .. seealso::
        - :doc:`jetstream.recvinterface.RecvInterface`

    Use ``api.server`` to access methods for common server operations.

    Example:
        .. code-block:: python

            api = JetStream().api()
            api.server.<method>()

    .. seealso::
        - :doc:`jetstream.serverinterface.ServerInterface`
    """

    def __init__(self, socket_interface, sync_interface, logging_interface, listen=False):
        self.__socket_interface = socket_interface
        self.__api_version = None
        self.send = SendInterface(self, sync_interface, logging_interface)
        self.recv = RecvInterface(self, sync_interface, logging_interface)
        self.server = ServerInterface(self, sync_interface, logging_interface)
        self.version = API_VERSION

        self._listen = listen

    @property
    def api_version(self):
        """The API version supported by the currently connected server.

        This is used by the API interfaces to determine whether they are compatible with the
        connected server. If a method is not compatible with the reported API version it will
        raise ``ApiVersionError``.
        """

        if not self.is_connected():
            raise NotConnectedError()

        return self.__api_version

    @property
    def serverAddress(self):
        """Returns the server address used for this connection.

        This property is read only. It must be assigned from :meth:`connect`.
        """
        return self.__socket_interface.serverAddress

    @property
    def serverPort(self):
        """Returns the server port used for this connection.

        This property is read only. It must be assigned from :meth:`connect`.
        """
        return self.__socket_interface.serverPort

    @property
    def timeoutSeconds(self):
        """Returns the timeout, in seconds, used to terminate blocking commands.

        This property is read only. It must be assigned from :meth:`connect`.

        Note:
            Timeout functionality is currently *unimplemented*.
        """
        return self.__socket_interface.timeoutSeconds

    def connect(self, serverAddress=NULL, serverPort=NULL, timeoutSeconds=NULL, certificatePath=NULL, sslCertificateRequirements=ssl.CERT_NONE):
        """Connects to a JetStream server.

        If this interface is already connected, calling this method will fail.

        Note:
            Timeout functionality is currently *unimplemented*.

        Args:
            serverAddress (str): The address (host name) of the server to connect to.
            serverPort (int): The port on which to connect.
            timeoutSeconds (float): A timeout, in seconds, used to terminate blocking commands. If unassigned, blocking commands will never unblock.
            certificatePath (str): Certificate of the JetStream server you are connecting to. Setting the ``certificatePath`` implies strict SSL certificate checks.
            sslCertificateRequirements (int): Level of strictness when checking SSL certificates. Can be one of: ``ssl.CERT_NONE`` (0), ``ssl.CERT_OPTIONAL`` (1), or ``ssl.CERT_REQUIRED`` (2).
        """
        self.__socket_interface.serverAddress = serverAddress
        self.__socket_interface.serverPort = serverPort
        self.__socket_interface.timeoutSeconds = timeoutSeconds
        self.__socket_interface.certificatePath = certificatePath
        self.__socket_interface.sslCertificateRequirements = sslCertificateRequirements
        self.__connect()

    def listen(self, listenAddress=NULL, listenPort=NULL, timeoutSeconds=NULL, certificatePath=NULL):
        """Listens for a connection from JetStream server.

        This is useful when JetStream server started with ``--api-connect`` argument. If this interface is already connected, calling this method will fail.

        Note:
            Timeout functionality is currently *unimplemented*.

        Args:
            listenAddress (str): The address (host name) to listen on for incoming connections.
            listenPort (int): The port on which to listen.
            timeoutSeconds (float): A timeout, in seconds, used to terminate blocking commands. If unassigned, blocking commands will never unblock.
            certificatePath (str): Certificate of the JetStream server you are connecting to.
        """
        self.__socket_interface.serverAddress = listenAddress
        self.__socket_interface.serverPort = listenPort
        self.__socket_interface.timeoutSeconds = timeoutSeconds
        self.__socket_interface.certificatePath = certificatePath
        self.__listen()

    def get_listen_socket(self, listenAddress=NULL, listenPort=NULL, timeoutSeconds=NULL, certificatePath=NULL):
        """Get a low level listening socket for API connections from JetStream server.

        This is useful when JetStream server started with ``--api-connect`` argument. If this interface is already connected, calling this method will fail.

        Args:
            listenAddress (str): The address (host name) to listen on for incoming connections.
            listenPort (int): The port on which to listen.
            timeoutSeconds (float): A timeout, in seconds, used to terminate blocking commands. If unassigned, blocking commands will never unblock.
            certificatePath (str): Certificate of the JetStream server you are connecting to.
        """
        self.__socket_interface.serverAddress = listenAddress
        self.__socket_interface.serverPort = listenPort
        self.__socket_interface.timeoutSeconds = timeoutSeconds
        self.__socket_interface.certificatePath = certificatePath
        return self.__socket_interface.get_listen_socket()

    def secure_socket(self, socket):
        """Wrap the given socket in SSL. 

        This is useful for securing sockets when listening for incoming connections with :func:`~jetstream.apiinterface.APIInterface.get_listen_socket`.

        Args:
            socket (socketobject): The socket to secure.
        """
        return self.__socket_interface.secure_socket(socket)

    def disconnect(self):
        """Immediately disconnects this interface from the JetStream server.

        If this interface is not connected to a JetStream server, this does nothing.
        """ 
        self.__socket_interface.disconnect()
        self.__api_version = None
        self.send.api_version = None
        self.recv.api_version = None
        self.server.api_version = None

    def is_connected(self):
        """Returns True if this API interface is connected; else returns False."""
        return self.__socket_interface.is_connected()

    def is_secured(self):
        """Returns True if this API interface is secured; else returns False."""
        return self.__socket_interface.is_secured()

    def get_cipher(self):
        """
        Returns a three-value tuple containing the name of the cipher being used, 
        the version of the SSL protocol that defines its use, and the number of 
        secret bits being used. If no secure connection has been established, returns None.
        """
        return self.__socket_interface.get_cipher()

    def get_server_certificate(self):
        """Returns server’s certificate as a PEM-encoded string if secure connection is established; else returns None"""
        return self.__socket_interface.get_server_certificate()

    def get_server_certificate_info(self):
        """Returns server’s certificate info as a dict if secure connection is established; else returns None"""
        return self.__socket_interface.get_server_certificate_info()

    def send_command(self, command, options=None):
        """Sends a raw API command to the server, returning the response"""
        return self._send(command, options)

    def __enter__(self):
        if self._listen:
            self.__listen()
        else:
            self.__connect()

        return self

    def __exit__(self, exc_type, exc_value, exc_tb):
        self.disconnect()

    def _send(self, command, options=None):
        packet = self.__socket_interface.create_packet(command, options)
        return self.__socket_interface.send_packet(packet)

    def __connect(self):
        self.__socket_interface.connect()
        self.__set_version()

    def __listen(self):
        self.__socket_interface.listen()
        self.__set_version()

    def __set_version(self):
        """Each time this interface connects it issues a 'getApiVersion' command to
        the server and stores the result. Each API methods wrapped checks its
        compatibility against this version.
        """
        try:
            result = self._send('getServerInfo')
            self.__api_version = result.get('apiVersion', None)
            self.send.api_version = self.__api_version
            self.recv.api_version = self.__api_version
            self.server.api_version = self.__api_version
        except JetStreamClientException:
            # Server does not support getApiVersion(). API version filter is disabled.
            self.__api_version = None
        except AttributeError:
            # Result is not expected type. API version filter is disabled.
            self.__api_version = None
