#!/usr/bin/env python
# encoding: utf-8
#
# Copyright 2017 Jet Digital Inc. All rights reserved.
#------------------------------------------------------------------------------------------------
# pylint: disable=unnecessary-pass

"""
Exception Hierarchy
-------------------

| :any:`JetStreamClientException`
| ├── :any:`CommandError`
|     ├── :any:`APIMethodNotFoundError`
|     ├── :any:`APIMethodDeprecatedError`
|     ├── :any:`APIConnectionUnavailableError`
|     ├── :any:`DestinationNotFoundError`
|     ├── :any:`ConnectionNotFoundError`
|     ├── :any:`FileSystemError`
|     ├── :any:`FileSystemOperationNotSupportedError`
|     ├── :any:`InsufficientPermissionsError`
|     ├── :any:`InvalidCredentialsError`
|     ├── :any:`InvalidLicenseError`
|     ├── :any:`InvalidParameterFormatError`
|     ├── :any:`InvalidParameterError`
|     ├── :any:`ManifestNotFoundError`
|     ├── :any:`NotAuthenticatedError`
|     ├── :any:`OperationFailedError`
|     ├── :any:`RequestNotFoundError`
|     ├── :any:`ResourceUnavailableError`
|     ├── :any:`SharedTokenNotFoundError`
|     └── :any:`TransferNotFoundError`
| ├── :any:`NotConnectedError`
| ├── :any:`ProtocolError`
|     ├── :any:`ApiVersionError`
|     └── :any:`PacketTooBigError`
| └── :any:`UnsupportedSSLVersionError`
| """

class JetStreamClientException(Exception):
    "``JetStreamClientException`` is the base class for all JetStream API exceptions."
    pass

class NotConnectedError(JetStreamClientException):
    """``NotConnectedError`` is raised when an attempt is made to send a command from an API interface
    which is not connected to a JetStream server."""
    pass

class ProtocolError(JetStreamClientException):
    """``ProtocolError`` is raised when a transmission error between the API and the JetStream
    server resulted in a command which was not recognized, or did not match the expected format."""
    pass

class ApiVersionError(ProtocolError):
    """``ApiVersionError`` is raised when an API command is used whose interface is not compatible
    with the connected JetStream server."""
    def __init__(self, name, requires_version, connected_version):
        super(ApiVersionError, self).__init__()
        self.name = name
        self.requires_version = requires_version
        self.connected_version = connected_version

    def __str__(self):
        return '{0} requires API version {1}; the connected server is API version {2}'.format(self.name, self.requires_version, self.connected_version)

class PacketTooBigError(ProtocolError):
    """``PacketTooBigError`` is raised when transmission error between the API and the JetStream
    server resulted in a packet which overflowed the maximum size allowed."""
    pass

class UnsupportedSSLVersionError(JetStreamClientException):
    """``UnsupportedSSLVersionError`` is raised when the client doesn't support the minimum SSL version
    required by the server"""
    pass

class CommandError(JetStreamClientException):
    """``CommandError`` is a catch-all exception for API command failures without more a specific error type."""
    pass

class APIMethodNotFoundError(CommandError):
    """``APIMethodNotFoundError`` is raised when the server does not recognize an API function."""
    pass

class APIMethodDeprecatedError(CommandError):
    """``APIMethodDeprecatedError`` is raised when an API function no longer exists."""
    pass

class APIConnectionUnavailableError(CommandError):
    """``APIConnectionUnavailableError`` may be raised before the server shuts down the API connection, for
    instance, while the server is terminating."""
    pass

class InvalidParameterFormatError(CommandError):
    """``InvalidParameterFormatError`` is raised when a parameter to an API function is missing, is the wrong
    type, or generally is not structured correctly."""
    pass

class InvalidParameterError(CommandError):
    """``InvalidParameterError`` is raised when a parameter to an API function is correctly structured, but has
    invalid data, such as a number that it our of range or a string that has an unexpected format."""
    pass

class InvalidCredentialsError(CommandError):
    """``InvalidCredentialsError`` is raised when invalid credentials, such as passwords, are used to
    authenticate the user."""
    pass

class NotAuthenticatedError(CommandError):
    """``NotAuthenticatedError`` is raised when a restricted API function is called by a user who has not authenticated."""
    pass

class InsufficientPermissionsError(CommandError):
    """``InsufficientPermissionsError`` is raised when an authenticated user does not have permission to complete
    the operation. For instance, some functions require superuser privileges."""
    pass

class FileSystemError(CommandError):
    """``FileSystemError`` is raised when there us an I/O error."""
    pass

class FileSystemOperationNotSupportedError(CommandError):
    """``FileSystemOperationNotSupported`` is raised when the requested operation is no supported on the
    underlying file system."""
    pass

class ResourceUnavailableError(CommandError):
    """``ResourceUnavailableError`` is raised when a hardware or OS resource, such as memory, cannot be allocated."""
    pass

class OperationFailedError(CommandError):
    """``OperationFailedError`` is a general exception when the API function cannot be completed."""
    pass

class ManifestNotFoundError(CommandError):
    """``ManifestNotFoundError`` is raised when a manifest id cannot be accessed by the user."""
    pass

class DestinationNotFoundError(CommandError):
    """``DestinationNotFoundError`` is raised when a destination id cannot be accessed by the user."""
    pass

class ConnectionNotFoundError(CommandError):
    """``ConnectionNotFoundError`` is raised when a connection id cannot be accessed by the user."""
    pass

class TransferNotFoundError(CommandError):
    """``TransferNotFoundError`` is raised when a transfer id cannot be accessed by the user."""
    pass

class SharedTokenNotFoundError(CommandError):
    """``SharedTokenNotFoundError`` is raised when a shared token cannot be accessed by the user."""
    pass

class RequestNotFoundError(CommandError):
    """``RequestNotFoundError`` is raised when a request id cannot be accessed by the user."""
    pass

class InvalidLicenseError(CommandError):
    """``InvalidLicenseError`` is raised when the server does not have a valid license but the operation requires one."""
    pass

# [backcompat]
JetstreamClientException = JetStreamClientException
