#!/usr/bin/env python
# encoding: utf-8
#
# Copyright 2017 Jet Digital Inc. All rights reserved.
#------------------------------------------------------------------------------------------------
"""
The Helper module contains convenience classes that may be used to assist with the primary
API classes.
"""

import os

from jetstream.constants import TRANSFER_FLAG_READ, TRANSFER_FLAG_WRITE

class Helper(object):
    """
    Args:
        api: A JetStream API interface, as constructed from :meth:`~.jetstream.api.JetStream.api`. This may also be None if using methods that do not require an API connection.
    """

    def __init__(self, api):
        self.api = api

    @staticmethod
    def getFileMappings(fileList, relative_to=None):
        """Builds a file mapping from a list of paths.

        The :meth:`~.jetstream.sendinterface.Send.createManifest` command requires a file mapping which
        provides one or more source paths, with a target path specified for each. In the case where
        you are satisfied to replicate the source structure at the destination, the mapping is
        trivially a 1:1 association. This helper method builds a file mapping of this type.

        *This method does not require an API connection.*

        Args:
            fileList (tuple, list): A sequence of file names or paths.
            relative_to (str): (optional) If specified, transforms all target paths to be
                relative to this path. Since the source paths must be absolute paths, this
                allows the file mapping to maintain a relative directory structure, but without
                replicating the full directory tree at the receiver.

        Returns:
            dict: A dictionary that maps each source file name/path to an equivalent path at the receiver.
        """
        if not isinstance(fileList, (tuple, list)):
            raise TypeError(type(fileList))

        if relative_to:
            make_relative = lambda p : os.path.relpath(p, relative_to)
        else:
            make_relative = lambda p : p

        return dict((path, make_relative(path)) for path in fileList)

    def getFilteredManifestFiles(self, predicate):
        """Applies the supplied filter to all manifests from the connected API, and returns
        a list of manifests, each with a list of file mappings that satisfy the filter.

        *This method requires an API connection.*

        Args:
            predicate: A callable that accepts a :ref:`structure-file-mapping`, and returns True if
                the file's manifest should be included in the results.

        Returns:
            list: Returns a list of :ref:`structure-manifest` items. Each manifest will supplemented with a ``fileMappings`` key which contains a list of :ref:`structure-file-mapping` items for all files matching the specified filter.
        """
        manifests = []
        for manifest in self.api.send.getManifests():
            manifestFiles = self.api.send.getManifestFiles(manifest)
            files = [f for f in manifestFiles['fileMappings'] if (predicate(f))]
            if files:
                manifest['fileMappings'] = files
                manifests.append(manifest)
        return manifests

    @staticmethod
    def getTransferFlags(readEnabled=True, writeEnabled=True):
        """Returns the transferFlags appropriate for the desired read/write behavior.

        The :meth:`~.jetstream.sendinterface.Send.createTransfer` command requires a ``transferFlags``
        value to specify whether reads and/or writes are enabled for the transfer.

        This method creates the ``transferFlags`` value for you.

        *This method does not require an API connection.*

        Args:
            readEnabled (bool): If True, enable reads in the transfer flags.
            writeEnabled (bool): If True, enable writes in the transfer flags.

        Returns:
            int: Returns a value appropriate for the ``transferFlags`` parameter
        """
        return (TRANSFER_FLAG_READ if readEnabled else 0) | (TRANSFER_FLAG_WRITE if writeEnabled else 0)

    def sendFiles(self, destinationId, sourcePath, destinationPath=None, userName=None, password=None):
        """Starts a file transfer and returns a TransferObject.

        This starts a file transfer for the specified path, to the specified destination.
        A new manifest is created for the path, and a new transfer is started for this manifest.

        A :class:`TransferObject` is returned, with which the transfer's progress can be monitored.

        *This method requires an API connection.*

        Args:
            destinationId (str, dict): The destination ID, as returned from :meth:`createDestination`.
                Also allowed is the full dict from :meth:`createDestination`. The destination ID
                is extracted from the ``destinationId`` key.
            sourcePath: The path to the source file or directory. This file/directory will be copied to the destination.
            destinationPath: (optional) Specifies a directory that will be created as the parent
                for all files in the transfer. The manifest paths are respected, and added below
                this destination path.
            userName: (optional) User name to authenticate the transfer. The transfer is authenticated by the receiving server.
            password: (optional) Password to authenticate the transfer. The transfer is authenticated by the receiving server.

        .. tip::
            Use ``getpass.getuser()`` to auto-fill the ``userName`` parameter.

            Use ``getpass.getpass()`` to prompt for the ``password`` parameter interactively.

        Returns:
            obj: :class:`TransferObject`
        """

        base_name = os.path.basename(sourcePath)
        manifest = self.api.send.createManifest({sourcePath: base_name})
        manifest = self.api.send.waitForManifest(manifest)
        transfer = self.api.send.createTransfer(destinationId, manifest['manifestId'], destinationPath=destinationPath, userName=userName, password=password)
        return TransferObject(self.api, destinationId, transfer['transferId'])


class TransferObject(object):
    """
    Args:
        api: A JetStream API interface, as constructed from :meth:`~.jetstream.api.JetStream.api`.
             This must be a connected and authenticated API for the server that is sending the transfer.
        destinationId: The ``DestinationId`` (as returned from :meth:`~.jetstream.sendinterface.SendInterface.createDestination`)
                       to which the transfer is being sent.
        transferId: The ``TransferId`` (as returned from :meth:`~.jetstream.sendinterface.SendInterface.createTransfer`)
                       for the transfer.
    """

    def __init__(self, api, destinationId, transferId):
        self.api = api
        self.destinationId = destinationId
        self.transferId = transferId

    def getProgress(self):
        """
        Returns a percentage (0 - 100) representing the current progress for the transfer.

        This value alone does not necessarily reflect whether the transfer has finished, or has failed. Compare :meth:`getStatus`.

        Returns:
            float: A percentage (0.0 - 100.0) representing the current progress for the transfer.
        """
        stat = self.getRawStats()
        return stat['blocksAcked'] * 100.0 / stat['totalBlocks']

    def getStatus(self):
        """
        Returns a string describing the current status for the transfer. This string represents the "status", extracted from the :ref:`structure-transfer`. Compare :meth:`getRawStats()`.

        The status "complete" indicates that the transfer has finished successfully.

        The status "error" indicates that the transfer has failed. Use :meth:`getRawStats()` to retrieve the full :ref:`structure-transfer`. Its ``errorMessage`` will describe the cause of the error.

        Other strings may be used to represent the progression of the transfer.

        Returns:
            string: A string describing the current status for the transfer.
        """
        return self.getRawStats()['status']

    def getRawStats(self):
        """
        Returns a :ref:`structure-transfer` containing the full description for the transfer.

        Returns:
            dict: :ref:`structure-transfer` containing the full description for the transfer.
        """
        return self.api.send.getTransfer(transferId=self.transferId)
