#!/usr/bin/env python
# encoding: utf-8
#
# Copyright 2017 Jet Digital Inc. All rights reserved.
#------------------------------------------------------------------------------------------------

from jetstream.apidecorator import jetapi

#------------------------------------------------------------------------------------------------
class RecvInterface(object):
    # pylint: disable=protected-access
    """``RecvInterface`` provides methods for controlling aspects of JetStream related to receiving
    files. ``RecvInterface`` is accessed from the :class:`~.jetstream.apiinterface.APIInterface`.
    """

    def __init__(self, api_interface, sync_interface, _logging_interface):
        self._api_interface = api_interface
        self.api_version = None
        self.getSources = sync_interface.create_sync_fn(api_interface, self.getSourcesAsync)
        self.getTransfers = sync_interface.create_sync_fn(api_interface, self.getTransfersAsync)
        self.allowOnlyAuthorizedClients = sync_interface.create_sync_fn(api_interface, self.allowOnlyAuthorizedClientsAsync)
        self.addAuthorizedClient = sync_interface.create_sync_fn(api_interface, self.addAuthorizedClientAsync)
        self.removeAuthorizedClient = sync_interface.create_sync_fn(api_interface, self.removeAuthorizedClientAsync)

    @jetapi(min_api=2, auth=True)
    def getSourcesAsync(self):
        """Submits a request to retrieve the sources that have established a connection with this Receiver.

        A "source" is a sending JetStream, from the viewpoint of the receiver.

        Returns:
            dict: Returns a :ref:`structure-request` containing a ``requestId`` for this request. Use :meth:`~.jetstream.serverinterface.ServerInterface.getRequestStatus` or :meth:`~.jetstream.serverinterface.ServerInterface.getRequestResult` to monitor the status of the request.

            :meth:`~.jetstream.serverinterface.ServerInterface.getRequestResult` will return dictionary :ref:`structure-source`.
        """
        return self._api_interface._send('recv.getSourcesAsync')

    @jetapi(min_api=4, auth=True, getTransferInfo=bool)
    def getTransfersAsync(self, getTransferInfo=False):
        """Submits a request to retrieve the transfers from a receive server.

        This command differs from :meth:`~.jetstream.sendinterface.SendInterface.getTransfers` in
        that this result describes the transfers being received on a server. The `SendInterface`
        method describes transfers being sent from a server.

        Returns:
            dict: Returns a :ref:`structure-request` containing a ``requestId`` for this request. Use :meth:`~.jetstream.serverinterface.ServerInterface.getRequestStatus` or :meth:`~.jetstream.serverinterface.ServerInterface.getRequestResult` to monitor the status of the request.

            :meth:`~.jetstream.serverinterface.ServerInterface.getRequestResult` will return dictionary :ref:`structure-recv-transfer`.
        """
        return self._api_interface._send('recv.getTransfersAsync', {'getTransferInfo': getTransferInfo})

    @jetapi(min_api=7, auth=True, superuser=True, onlyAuthorizedClientsAllowed=bool)
    def allowOnlyAuthorizedClientsAsync(self, onlyAuthorizedClientsAllowed):
        """Submits a request to enable or disable the restriction to allow only authorized clients to connect to this server.

        When enabled, only incoming connections from trusted clients will be accepted.
        Trusted clients can be added using :meth:`addAuthorizedClientAsync` and removed
        using :meth:`removeAuthorizedClientAsync`.

        Args:
            onlyAuthorizedClientsAllowed (bool): ``True`` restricts connections to
                allow only clients authorized with :meth:`addAuthorizedClientAsync`.
                ``False`` allows connections from any client. Also allowed is a dict
                with a bool value extracted from the ``onlyAuthorizedClientsAllowed`` key.

        Returns:
            dict: Returns a :ref:`structure-request` containing a ``requestId`` for this request. Use :meth:`~.jetstream.serverinterface.ServerInterface.getRequestStatus` or :meth:`~.jetstream.serverinterface.ServerInterface.getRequestResult` to monitor the status of the request.
        """
        return self._api_interface._send('recv.allowOnlyAuthorizedClientsAsync', {'onlyAuthorizedClientsAllowed': onlyAuthorizedClientsAllowed})

    @jetapi(min_api=7, auth=True, superuser=True, clientPublicKey=str)
    def addAuthorizedClientAsync(self, clientPublicKey):
        """Adds a public key to the list of authorized clients.

        If :meth:`allowOnlyAuthorizedClientsAsync` is used to restrict incoming connections,
        a Receiver server will allow incoming connections only from trusted clients. Adding a
        client's public key will permit it to connect.

        If a JetStream server was launched with the ``--persistent-state-dir`` flag, its list of
        authorized clients will be saved, and restored if the server is restarted. If a
        JetStream server is launched without this flag, its list of authorized clients must be
        configured again.

        Args:
            clientPublicKey (str): The public key of an trusted client that is authorized to
                connect to a Receiver server. Here, "client" refers to a JetStream Sender server.
                Also allowed is a dict with a string value extracted from the ``clientPublicKey`` key.

        Returns:
            dict: Returns a :ref:`structure-request` containing a ``requestId`` for this request. Use :meth:`~.jetstream.serverinterface.ServerInterface.getRequestStatus` or :meth:`~.jetstream.serverinterface.ServerInterface.getRequestResult` to monitor the status of the request.
        """
        return self._api_interface._send('recv.addAuthorizedClientAsync', {'clientPublicKey': clientPublicKey})

    @jetapi(min_api=7, auth=True, superuser=True, clientPublicKey=str)
    def removeAuthorizedClientAsync(self, clientPublicKey):
        """Removes a public key from the list of authorized clients.

        If :meth:`allowOnlyAuthorizedClientsAsync` is used to restrict incoming connections,
        a Receiver server will allow incoming connections only from trusted clients. Removing a
        client's public key will result in its connections being refused.

        If a JetStream server was launched with the ``--persistent-state-dir`` flag, its list of
        authorized clients will be saved, and restored if the server is restarted. If a
        JetStream server is launched without this flag, its list of authorized clients must be
        configured again.

        Args:
            clientPublicKey (str): The public key of an untrusted client that is not authorized
                to connect to a Receiver server. Here, "client" refers to a JetStream Sender server.
                Also allowed is a dict with a string value extracted from the ``clientPublicKey`` key.

        Returns:
            dict: Returns a :ref:`structure-request` containing a ``requestId`` for this request. Use :meth:`~.jetstream.serverinterface.ServerInterface.getRequestStatus` or :meth:`~.jetstream.serverinterface.ServerInterface.getRequestResult` to monitor the status of the request.
        """
        return self._api_interface._send('recv.removeAuthorizedClientAsync', {'clientPublicKey': clientPublicKey})
