#!/usr/bin/env python
# encoding: utf-8
#
# Copyright 2017 Jet Digital Inc. All rights reserved.
#------------------------------------------------------------------------------------------------
from jetstream.global_constants import Constants

import sys
if sys.version_info[0] > 2:
    longint=int
else:
    longint=long # pylint: disable=undefined-variable

from jetstream.apidecorator import jetapi, jetapi_extract_param

#------------------------------------------------------------------------------------------------
class SendInterface(object):
    # pylint: disable=protected-access
    """``SendInterface`` provides methods for controlling aspects of JetStream related to sending
    files. ``SendInterface`` is accessed from the :class:`~.apiinterface.APIInterface`.

    .. tip::
        Methods whose names end with ``Async`` are asynchronous commands. These commands are sent
        to the server as a *request* and queued for processing. Asynchronous methods return a
        ``RequestId``. Use :meth:`~.jetstream.serverinterface.ServerInterface.getRequestStatus` and
        :meth:`~.jetstream.serverinterface.ServerInterface.getRequestResult` to monitor the
        status of a request.
    """

    def __init__(self, api_interface, sync_interface, logging_interface):
        self._api_interface = api_interface
        self.__logging_interface = logging_interface
        self.api_version = None
        self.resumeTransfer = sync_interface.create_sync_fn(api_interface, self.resumeTransferAsync)
        self.setSendRateMax = sync_interface.create_sync_fn(api_interface, self.setSendRateMaxAsync)
        self.setPriorityLanes = sync_interface.create_sync_fn(api_interface, self.setPriorityLanesAsync)
        self.suspendTransfer = sync_interface.create_sync_fn(api_interface, self.suspendTransferAsync)
        self.setTransferPriority = sync_interface.create_sync_fn(api_interface, self.setTransferPriorityAsync)
        self.setTransferSendRateMax = sync_interface.create_sync_fn(api_interface, self.setTransferSendRateMaxAsync)
        self.updateConnectionParams = sync_interface.create_sync_fn(api_interface, self.updateConnectionParamsAsync)
        self.schedulePing = sync_interface.create_sync_fn(api_interface, self.schedulePingAsync)

    # -- -- -- -- -- -- -- -- -- -- -- -- -- -- -- -- -- -- -- -- -- -- -- -- -- -- -- -- -- --
    # Destination Configuration

    @jetapi(min_api=2, auth=True, destinationId=str, sendRateMax=float)
    def setSendRateMaxAsync(self, destinationId, sendRateMax):
        """Submits a request to adjust the maximum send rate for the specified destination.

        When the maximum send rate is adjusted, the minimum send rate and actual send rate will also
        be adjusted, if necessary, so as not to exceed the maximum rate.

        Args:
            destinationId (str, dict): A destination ID or a :ref:`structure-destination`.
            sendRateMax (float): The maximum send rate that will be allowed by the JetStream rate controller, in kilobits/second.

        Returns:
            dict: Returns a :ref:`structure-request` containing a ``requestId`` for this request. Use :meth:`~.jetstream.serverinterface.ServerInterface.getRequestStatus` or :meth:`~.jetstream.serverinterface.ServerInterface.getRequestResult` to monitor the status of the request.

        Example:
            >>> request = send.setSendRateAsync(destinationId, sendRate=750000)
            {'requestId': 1}
            >>> result = send.getRequestResult(request)
            {'sendRateMax': 750000.0}
        """
        return self._api_interface._send('send.setSendRateMaxAsync', {'destinationId': destinationId, 'sendRateMax': sendRateMax})

    @jetapi(min_api=23, auth=True, destinationId=str)
    def setPriorityLanesAsync(self, destinationId, priorityLanes):
        """Set the priority lanes for transfers.

        Each priority lane represents a priority queue of transfers. Transfers in each lane
        are transmitted simultaneously using the given proportion of the current send rate.

        Args:
            destinationId (str, dict): A destination ID or a :ref:`structure-destination`.
            priorityLanes (dict): A mapping from lane name to lane rate proportions. See below for rules when specifying custom lanes.

        Returns:
            dict: Returns a :ref:`structure-request` containing a ``requestId`` for this request. Use :meth:`~.jetstream.serverinterface.ServerInterface.getRequestStatus` or :meth:`~.jetstream.serverinterface.ServerInterface.getRequestResult` to monitor the status of the request.

        Raises:
            CommandError: The destinationId was not found or the priority lanes are invalid.

        ``priorityLanes`` mapping has to be specified according to the following rules:
            - All lanes must be updated simulataneously.
            - There must be at least one lane.
            - Lane names cannot be blank.
            - Lane names must be unique.
            - Lane rate proportions must sum to 1.0.
            - At most one lane can have a rate proportion of 0.0, which is the idle lane: data will only be transfered when all other lanes are empty.
            - The number of lanes cannot be changed if there are transfers in-progress.
            - The name of a lane cannot be changed if the lane has a transfer in-progress.
            - Rates may be updated while transfer are in-progress.

        By default, the priority lanes are set as follows

        ======  ====================
        Lane    Bandwidth allocation
        ======  ====================
        High    70%                 
        Normal  25%                 
        Low     5%                  
        Idle    0%                  
        ======  ====================

        .. versionadded:: 1.6.1
        """
        return self._api_interface._send('send.setPriorityLanesAsync', {'destinationId': destinationId, 'priorityLanes': priorityLanes})

    # -- -- -- -- -- -- -- -- -- -- -- -- -- -- -- -- -- -- -- -- -- -- -- -- -- -- -- -- -- --
    # Destination

    @jetapi(min_api=2, auth=True, destinationAddress=str, destinationPort=int, sendRateMax=float, authenticateDestination=bool, destinationPublicKey=str, relayAddress=str, relayPort=int, senderPort=int, exclusive=bool, permanent=bool)
    def createDestination(self, destinationAddress, sendRateMax=None, destinationPort=None, authenticateDestination=None, destinationPublicKey=None, relayAddress=None, relayPort=None, senderPort=None, exclusive=None, permanent=None):
        """Adds a new destination to the JetStream server or returns an existing destination with matching parameters.

        A destination describes to JetStream how to configure its send operation such that its
        data will be successfully retrieved by a JetStream receiver. Fundamentally, a destination
        is identified by its host name (or IP) and its receiver port.

        Note:
            Creating a destination does not guarantee there is a listening JetStream server at
            the other end.

        Args:
            destinationAddress (str): The address (host name or IP) through which to connect to the destination.
            sendRateMax (int): The maximum send rate allowed for this server, in kilobits-per-second. If the destination already exists, then this value is ignored.
            destinationPort (int): The receiver UDP port for a destination.
            authenticateDestination (bool): Only allow connections to a destination with the public key specified by `destinationPublicKey`.
            destinationPublicKey (str): The public key of the destination to use for authentication.
            relayAddress (str): The address (host name or IP) for the relay server, if using a relay. Otherwise unspecified.
            relayPort (int): The port for the relay server, if using a relay. Otherwise unspecified.
            senderPort (int): The UDP port of the JetStream Sender server. Not specified or 0 will choose a new port. -1 will multiplex the receiver port, if possible.
            exclusive (bool): Create a new destination, even there is an existing one with matching parameters.
            permanent (bool): Do not allow the destination to be garbage collected once all transfers are deleted.

        Returns:
            dict: :ref:`structure-destination`

        Example:

            >>> sender.createDestination(destinationAddress='receiver01')

        .. versionchanged:: 1.7.0 Removed ``sendRateMin``, ``sendRate``, ``encryptionEnabled`` and ``tag`` parameters.
        .. versionchanged:: 1.7.1 Added ``exclusive`` parameter.
        .. versionchanged:: 2.3.0 Added ``permanent`` parameter.
        """
        return self._api_interface._send('send.createDestination', {'destinationAddress': destinationAddress, 'destinationPort': destinationPort, 'senderPort': senderPort, 'sendRateMax': sendRateMax, 'authenticateDestination': authenticateDestination, 'destinationPublicKey': destinationPublicKey, 'relayAddress': relayAddress, 'relayPort': relayPort, 'exclusive': exclusive, 'permanent': permanent})

    @jetapi(min_api=2, auth=True, destinationId=str)
    def getDestination(self, destinationId):
        """Returns the configuration for a specified destination.

        Args:
            destinationId (str, dict): A destination ID or a :ref:`structure-destination`.

        Returns:
            dict: :ref:`structure-destination`

        Raises:
            CommandError: The destination was not found.
        """
        return self._api_interface._send('send.getDestination', {'destinationId': destinationId})

    @jetapi(min_api=2, auth=True, limit=int)
    def getDestinations(self, limit=0):
        """Returns the current configurations for all known destinations.

        Args:
            limit (int): (optional) The maximum number of results to retrieve.
                         If not specified, all available transfers are returned.

        Returns:
            list: A list of :ref:`destination structures <structure-destination>`, each describing a destination.
        """
        return self._api_interface._send('send.getDestinations', {'limit': limit})

    @jetapi(min_api=2, auth=True, destinationId=str)
    def deleteDestination(self, destinationId):
        """Deletes the specified destination.

        This deletes all transfers associated with the destination. If the destination is recreated,
        any transfers will need to be recreated, as well.

        Note:
            A destination may not be deleted while any transfers are active. You must delete any
            active transfers first.

        Args:
            destinationId (str, dict): A destination ID or a :ref:`structure-destination`.

        Returns:
            None

        Raises:
            CommandError: The destination was not found.
        """
        return self._api_interface._send('send.deleteDestination', {'destinationId': destinationId})

    # -- -- -- -- -- -- -- -- -- -- -- -- -- -- -- -- -- -- -- -- -- -- -- -- -- -- -- -- -- --
    # Manifest

    @jetapi(min_api=2, auth=True, symlinkOptions=int)
    def createManifest(self, fileMappings=None, mappings=None, symlinkOptions=None):
        """Creates a manifest for the provided file mapping.

        The creation of a manifest happens immediately. The :meth:`createManifest` method
        returns a structure with ``manifestId``, which identifies the new manifest. This ``manifestId``
        indicates that the JetStream server has *started* to collect information from the
        file system for all of the specified files. However, the manifest is not necessarily
        yet ready for use to create a transfer. If one or more directory paths are
        specified for the manifest, or a large list of files, then the server must
        recursively visit each subdirectory and file, which may take some time. You
        will need to monitor the progress of the manifest to know when the traversal has
        completed. Use :meth:`getManifest` or :meth:`waitForManifest` to monitor the status
        of a manifest.

        Args:
            fileMappings (dict): **Deprecated.** A dictionary mapping source paths to their desired path file name at the destination.
            mappings (dict): A dictionary mapping destination paths to source paths.
            symlinkOptions (int): A bitflag specifying how and whether to follow and copy symlinks during a transfer. If not specified, the default is to copy the contents of file or directory symlinks.

        Returns:
            dict: :ref:`structure-manifest`

        ``mappings``:
            A dictionary that maps the desired destination path to a given source path. For
            example, suppose you want to send the file "readme.txt". At your location (as the
            sender), the file exists at the path "/usr/docs". At its destination, you want the file
            to be named "introduction.txt" as part of a package below the path "project". The file
            mapping would be as follows:

            Each source path may be a file or directory and must exist at the time of the call.

            Example:

                >>> createManifest( mappings={'/project/introduction.txt': '/usr/docs/readme.txt'} )
        
            When the source files do not exist, or the file sizes are known in advance, a manifest that skips
            all I/O may be created (i.e. no directory traversal or existence checks are run). To do this, 
            specify a mapping with information about the source data:

                * ``src``: Source file name (it must exist when transfer is started).
                * ``size``: Size of the source file, specified in bytes.
                * ``flags``: Specifies whether the entry is a file or directory. Use one of the ``FILETYPE_*`` constants (See :data:`~jetstream.constants`).
                * ``access``: A string which represents the file permissions. The only recognized access is 'x' (executable).
                * ``exclude``: An array of exclude patterns in the file names i.e. ['*.png', '*.jpeg'] to exclude them from manifest
                * ``include``: An array of include patterns in the file name i.e. ['*.mp3', '*.mov'] to include them in the manifest.

            If neither ``size`` nor ``flags`` is specified, the entry is interpreted to be an empty directory. If ``flags`` specifies a directory, ``size`` is ignored.
            
            Example:

                >>> createManifest( mappings={'/storage/destinationFile.txt': {'src':'/sourceFile.txt', 'size':100}} )

                >>> createManifest( mappings={'/project/data/destinationEmptyDir': {'src':'/emptyDir'}} )

                >>> createManifest( mappings={'/path/to/destination': {'src':'/path/to/source', exclude: ['*.png', '*.PNG'], include: ['*.mp3']}} )


        ``symlinkOptions`` can be one of:
            * :data:`~jetstream.constants.SYMLINKS_IGNORE`
            * :data:`~jetstream.constants.SYMLINKS_COPY_CONTENTS` (default)

                >>> createManifest( mappings={'/project/': '/src/project/'}, symlinkOptions=SYMLINKS_IGNORE )

        .. versionchanged:: 1.8.2 Added ``symlinkOptions`` parameter. Added``flags`` and ``access`` to mappings.
        """
        return self._api_interface._send('send.createManifest', {'fileMappings': fileMappings, 'mappings': mappings, 'symlinkOptions': symlinkOptions})

    @jetapi(min_api=2, auth=True, manifestId=str)
    def getManifest(self, manifestId):
        """Returns a structure describing the current state of the given manifest.

        Args:
            manifestId (str, dict): A manifest ID or a :ref:`structure-manifest`.

        Returns:
            dict: :ref:`structure-manifest`

        Raises:
            CommandError: The manifest was not found.

        Tip:
            The ``status`` field in the returned structure may be used to determine whether the server
            is still processing the manifest. Please refer to :ref:`structure-manifest` for list of possible values and their meaning.
        """
        return self._api_interface._send('send.getManifest', {'manifestId': manifestId})

    @jetapi(min_api=2, auth=True, before=str, after=str, limit=int)
    def getManifests(self, before=None, after=None, limit=None):
        """Returns information for all known manifests.

        Args:
            before (str, dict): (optional) This expects a manifest ID. If specified,
                        filter results to include only manifests before this manifest ID.
                        The specified manifest ID will not be included in the result.
                        Also allowed is the full dict representating a manifest structure,
                        as from :meth:`createManifest`.
            after (str, dict): (optional) This expects a manifest ID. If specified,
                        filter results to include only manifests after this manifest ID.
                        The specified manifest ID will not be included in the result.
                        Also allowed is the full dict representating a manifest structure,
                        as from :meth:`createManifest`.
            limit (int): (optional) The maximum number of results to retrieve.
                         If not specified, all available manifests are returned.

        Returns:
            list: A list of :ref:`manifest structures <structure-manifest>`, each describing a manifest.
        """
        _after = jetapi_extract_param('manifestId', after)
        _before = jetapi_extract_param('manifestId', before)
        return self._api_interface._send('send.getManifests', {'before': _before, 'after': _after, 'limit': limit})

    @jetapi(min_api=2, auth=True, manifestId=str, limit=longint)
    def getManifestFiles(self, manifestId, limit=None):
        """Returns the file mappings for a manifest.

        Args:
            manifestId (str, dict): A manifest ID or a :ref:`structure-manifest`.

        Returns:
            dict: :ref:`structure-manifest-files`

        Raises:
            CommandError: The manifest was not found.
        """
        return self._api_interface._send('send.getManifestFiles', {'manifestId': manifestId, 'limit': limit})

    @jetapi(min_api=2, auth=True, manifestId=str)
    def deleteManifest(self, manifestId):
        """Deletes the specified manifest.

        A manifest may be deleted at any time, even if it has been used to initiate a
        transfer which is currently active.

        Args:
            manifestId (str, dict): A manifest ID or a :ref:`structure-manifest`.

        Returns:
            None

        Raises:
            CommandError: The manifest was not found.
        """
        return self._api_interface._send('send.deleteManifest', {'manifestId': manifestId})

    @jetapi(min_api=2, auth=True, manifestId=str)
    def waitForManifest(self, manifestId):
        """Waits for the server to finish processing a manifest.

        After a manifest is created, the server must recursively visit each subdirectory
        and file in the specified file mapping, which may take some time. This method
        may be used to wait until server has finished initializing the manifest.

        Note:
            This method will block until the manifest is complete.

        Args:
            manifestId (str, dict): A manifest ID or a :ref:`structure-manifest`.

        Returns:
            dict: :ref:`structure-manifest`

        Raises:
            CommandError: The manifest was not found.
        """
        return self._api_interface._send('send.waitForManifest', {'manifestId': manifestId})

    # -- -- -- -- -- -- -- -- -- -- -- -- -- -- -- -- -- -- -- -- -- -- -- -- -- -- -- -- -- --
    # Shared tokens

    @jetapi(min_api=14, auth=True, manifestId=str, password=str, downloadLimit=int, timeLimitSeconds=int, description=str)
    def createSharedToken(self, manifestId, password=None, downloadLimit=None, timeLimitSeconds=None, description=None):
        """Creates a shared token for a manifest.

        The returned structure provides a 'token' which may be be used by clients to log on and
        have access to the manifest.

        Args:
            manifestId (str, dict): A manifest ID or a :ref:`structure-manifest`.
            password (str): (optional) A password required to auth this token.
            downloadLimit: (optional) Number of times a transfer can be initiated using this token.
            timeLimitSeconds: (optional) Number of seconds from now before this token expires.
            description: (optional) A description of what this token represents.

        Returns:
            dict: :ref:`structure-token`

        Raises:
            CommandError: The manifest was not found.

        See Also:
            - :doc:`jetstream-server:using_jetstream/shared_links`

        .. versionadded:: 1.4.0
        """
        return self._api_interface._send('send.createSharedToken', {'manifestId': manifestId, 'downloadLimit': downloadLimit, 'timeLimitSeconds': timeLimitSeconds, 'password': password, 'description': description})

    @jetapi(min_api=14, auth=True, token=str)
    def getSharedToken(self, token):
        """Returns a structure describing a shared token.

        Args:
            token (str, dict): A Shared Token or :ref:`structure-token`.

        Returns:
            dict: :ref:`structure-token`

        Raises:
            CommandError: The shared token was not found.

        See Also:
            - :doc:`jetstream-server:using_jetstream/shared_links`

        .. versionadded:: 1.4.0
        """
        return self._api_interface._send('send.getSharedToken', {'token': token})

    @jetapi(min_api=14, auth=True, limit=int)
    def getSharedTokens(self, limit=None):
        """Returns information for all known shared tokens.

        Args:
            limit (int): (optional) The maximum number of results to retrieve. If not specified, all available shared tokens are returned.

        Returns:
            list: A list of :ref:`shared token structures <structure-token>`, each describing a shared token.

        See Also:
            - :doc:`jetstream-server:using_jetstream/shared_links`

        .. versionadded:: 1.4.0
        """
        return self._api_interface._send('send.getSharedTokens', {'limit': limit})

    @jetapi(min_api=14, auth=True, token=str)
    def deleteSharedToken(self, token):
        """Deletes the specified shared token.

        A shared token may be deleted at any time, even if it is currently being used
        as part of a transfer or log in session. The active transfer will continue and will not be terminated.

        Args:
            token (str, dict): A Shared Token or :ref:`structure-token`.

        Returns:
            None

        Raises:
            CommandError: The shared token was not found.

        .. versionadded:: 1.4.0
        """
        return self._api_interface._send('send.deleteSharedToken', {'token': token})

    @jetapi(min_api=34, auth=True, token=str, timeLimitSeconds=int, description=str)
    def updateSharedToken(self, token, timeLimitSeconds=None, description=None):
        """Updates the specified shared token.

        Args:
            token (str, dict): A Shared Token or :ref:`structure-token`.
            timeLimitSeconds: (optional) Number of seconds from now before this token expires.
            description: (optional) A description of what this token represents.

        Returns:
            None

        Raises:
            CommandError: The shared token was not found.

        .. versionadded:: 2.6.0
        """
        return self._api_interface._send('send.updateSharedToken', {'token': token, 'timeLimitSeconds': timeLimitSeconds, 'description': description})

    # -- -- -- -- -- -- -- -- -- -- -- -- -- -- -- -- -- -- -- -- -- -- -- -- -- -- -- -- -- --
    # Transfer

    @jetapi(min_api=2, auth=True, destinationId=str, manifestId=str, destinationPath=str, userName=str, password=str, priority=int, transferFlags=int, checkpointFrequencySeconds=int, sendRateMax=float, priorityLane=str, overwriteMode=int, minCipher=int, fileTimeResolutionMicroseconds=longint, timeoutSecond=float, token=str, transferLog=bool)
    def createTransfer(self, destinationId, manifestId, destinationPath=None, userName=None, password=None, priority=None, transferFlags=None, checkpointFrequencySeconds=None, sendRateMax=None, priorityLane=None, overwriteMode=None, minCipher=None, fileTimeResolutionMicroseconds=None, timeoutSeconds=None, userData=None, token=None, transferLog=False):
        """Initiates a new transfer from a manifest, to a given destination.

        The returned structure provides a ``transferId`` which may be submitted to :meth:`getTransfer()` in
        order to track the status of the transfer. Also provided is a ``processing`` key that
        indicates if the transfer is still in progress. The ``processing`` key clears to False
        once the transfer is completed, if it fails, or when it is suspended.

        Args:
            destinationId (str, dict): A destination ID or a :ref:`structure-destination`.
            manifestId (str, dict): A manifest ID or a :ref:`structure-manifest`.
            destinationPath: (optional) Specifies a directory that will be created as the parent \
                for all files in the transfer. The manifest paths are respected, and added below \
                this destination path.
            userName (str): (optional) User name to authenticate the transfer. The transfer is authenticated by the receiving server.
                Only one of ``userName`` and ``token`` should be used.
            token (str): (optional) Token used to authenticate the transfer. The transfer is authenticated by the receiving server.
                Only one of ``userName`` and ``token`` should be used.
            password (str): (optional) Password to authenticate the transfer. The transfer is authenticated by the receiving server.
            priority (int): (optional) Specifies the initial priority for the new transfer. \
                Lower values are considered higher priority; i.e. priority 1 takes precedence over \
                priority 2. Accepted values are between 0 and 4294967295, inclusive. The default \
                is 100. The priority of a transfer may be adjusted at any time using :meth:`setTransferPriorityAsync`.
            transferFlags (int): (optional) A flag which controls whether to enable writes and reads on the destination.
            checkpointFrequencySeconds (int): (optional) How frequently to take a checkpoint. \
                A checkpoint is a guarantee that files have been written to storage. If the \
                receiver of a transfer is terminated and subsequently resumed, the transfer \
                can safely continue from its last checkpoint. If 0 (zero), checkpoints are disabled. \
                If None, the transfer will use the default from the destination.
            sendRateMax (float): (optional) Specifies the maximum transfer rate in kilobits/second. The actual transfer rate \
                may be less depending on the destination settings and network conditions. The maximum send rate may \
                be adjusted at any time using :meth:`setTransferSendRateMaxAsync`.
            priorityLane (str): (optional) The priority lane for the new transfer.
            overwriteMode (int): (optional) How to handle creating files, if a file with the same name already exists.
            minCipher (int): (optional) Minimum desired encryption type.
            fileTimeResolutionMicroseconds (int): (optional) Files with modification times within the given number of microseconds are considered equal.
            timeoutSeconds (float): (optional) Fail the transfer if it's disconnected for this many seconds. 0 seconds indicates no timeout.
            userData (dict): (optional) A dictionary of string keys and string values used as the initial user data of the transfer.
            transferLog (bool): (optional) A flag which enable the creation of individual transfer log data. The server must have enabled the option --log-transfer-dir.

        Returns:
            dict: :ref:`structure-transfer`

        Raises:
            CommandError: The destination or manifest was not found.

        ``transferFlags`` can be one of:
            * :data:`~jetstream.constants.TRANSFER_FLAG_NONE`
            * :data:`~jetstream.constants.TRANSFER_FLAG_READ`
            * :data:`~jetstream.constants.TRANSFER_FLAG_WRITE`
            * :data:`~jetstream.constants.TRANSFER_FLAG_READ_WRITE`
            * :data:`~jetstream.constants.TRANSFER_FLAG_WRITE_MTIME`
            * :data:`~jetstream.constants.TRANSFER_FLAG_WRITE_ONLY_NEWER`

        ``overwriteMode`` can be one of:
            * :data:`~jetstream.constants.OVERWRITE_IN_PLACE_SAME_SIZE`
            * :data:`~jetstream.constants.OVERWRITE_NEVER`
            * :data:`~jetstream.constants.OVERWRITE_ALWAYS`
            * :data:`~jetstream.constants.OVERWRITE_VERSIONED`

        ``minCipher`` can be one of:
            * :data:`~jetstream.constants.MIN_CIPHER_NONE`
            * :data:`~jetstream.constants.MIN_CIPHER_UNENCRYPTED`
            * :data:`~jetstream.constants.MIN_CIPHER_ENCRYPTED_128`
            * :data:`~jetstream.constants.MIN_CIPHER_ENCRYPTED_256`

        Tip:
            Use ``getpass.getuser()`` to auto-fill the ``userName`` parameter.

            Use ``getpass.getpass()`` to prompt for the ``password`` parameter interactively.

        .. versionchanged:: 1.7.0 Added ``overwriteMode`` parameters.
        .. versionchanged:: 2.2.0 Added ``minCipher`` parameter.
        .. versionchanged:: 2.3.0 Added ``TRANSFER_FLAG_WRITE_MTIME`` and ``TRANSFER_FLAG_WRITE_ONLY_NEWER`` transferFlags. Renamed several overwrite modes. Added ``fileTimeResolutionMicroseconds``, ``timeoutSeconds`` and ``userData``.
        .. versionchanged:: 2.4.0 Added ``token`` parameter.
        """
        if userData is None:
            transferUserData = {'transferOrigin': str(Constants.transfer_origin_python)}
        else:
            transferUserData = userData.copy()
            transferUserData['transferOrigin'] = str(Constants.transfer_origin_python)

        if transferLog == True:
            transferUserData['transferLog'] = str(transferLog)
        
        if userName is None:
            userName = 'not provided - transfer will not start'

        return self._api_interface._send('send.createTransfer', {'destinationId': destinationId, 'manifestId': manifestId, 'destinationPath': destinationPath, 'userName': userName, 'token': token, 'password': password, 'priority': priority, 'transferFlags': transferFlags, 'checkpointFrequencySeconds': checkpointFrequencySeconds, 'sendRateMax': sendRateMax, 'priorityLane': priorityLane, 'overwriteMode': overwriteMode, 'minCipher': minCipher, 'fileTimeResolutionMicroseconds': fileTimeResolutionMicroseconds, 'timeoutSeconds': timeoutSeconds, 'userData': transferUserData})

    @jetapi(min_api=2, auth=True, transferId=str, getCompleteFlags=bool)
    def getTransfer(self, transferId, getCompleteFlags=None):
        """Returns a structure describing the current state of a transfer.

        Args:
            transferId (int, dict): A transfer ID or a :ref:`structure-transfer`.
            getCompleteFlags (bool): (optional) ``True`` request the transfer complete file flags, ``False`` don't request file flags.

        Returns:
            dict: :ref:`structure-transfer`

        Raises:
            CommandError: The transfer was not found.

        Tip:
            The ``status`` field in the returned structure may be used to determine the state of the transfer.
            Please refer to :ref:`structure-transfer` for list of possible values and their meaning.

        See Also:
            - :doc:`/api/timestamps`
        """
        return self._api_interface._send('send.getTransfer', {'transferId': transferId, 'getCompleteFlags': getCompleteFlags})

    @jetapi(min_api=2, auth=True, destinationId=str, before=str, after=str, limit=int)
    def getTransfers(self, destinationId=None, before=None, after=None, limit=None):
        """Returns a list of structures describing the state of all transfers.

        If the optional ``destinationId`` parameter is specified, the results are limited to only the transfers for the given destination. Otherwise, 
        all transfers are returned.

        Args:
            destinationId (str, dict): (optional) A destination ID or a :ref:`structure-destination`.
            before (str, dict): (optional) A transfer ID or a :ref:`structure-transfer`. If specified, \
                        filter results to include only transfers before this transferId. \
                        The specified transfer ID will not be included in the result.
            after (str, dict): (optional) A transfer ID or a :ref:`structure-transfer`. If specified, \
                        filter results to include only transfers after this transferId. \
                        The specified transfer ID will not be included in the result.
            limit (int): (optional) The maximum number of results to retrieve. \
                         If not specified, all available transfers are returned.

        Returns:
            list: A list of :ref:`transfer structures <structure-transfer>`, each describing a transfer.

        Raises:
            CommandError: The destination was not found.
        """
        _after = jetapi_extract_param('transferId', after)
        _before = jetapi_extract_param('transferId', before)
        return self._api_interface._send('send.getTransfers', {'destinationId': destinationId, 'before': _before, 'after': _after, 'limit': limit})

    @jetapi(min_api=2, auth=True, transferId=str)
    def resumeTransferAsync(self, transferId):
        """Resumes a transfer that has been suspended.

        Resuming a transfer will restore it to the server's processing queue,
        and it will resume sending data to its destination.

        Resuming a transfer moves the transfer back in
        the processing queue. The transfer may not resume sending files immediately,
        depending on the traffic from other ongoing transfers. Use :meth:`getTransfer`
        or :meth:`waitForTransfer` to monitor the status of the transfer.

        Only suspended transfers can be resumed. If a transfer is already complete,
        or has failed, this request will fail.

        Args:
            transferId (int, dict): A transfer ID or a :ref:`structure-transfer`.

        Returns:
            dict: :ref:`structure-request` containing a ``requestId`` for this request. Use :meth:`~.jetstream.serverinterface.ServerInterface.getRequestStatus` or :meth:`~.jetstream.serverinterface.ServerInterface.getRequestResult` to monitor the status of the request.

            :meth:`~.jetstream.serverinterface.ServerInterface.getRequestResult` will return :ref:`structure-transfer`.

        Raises:
            CommandError: The transfer was not found, or could not be resumed.

        Example:

            >>> request = send.resumeTransferAsync(transfer)
            {'requestId': 7}
            >>> status = send.getRequestStatus(request)
            {'status': 'complete', ...}
            >>> result = send.getRequestResult(request)
            {'transferId': 13}
            >>> transfer = send.getTransfer(result)
            {'status': 'resuming', ...}

        """
        return self._api_interface._send('send.resumeTransferAsync', {'transferId': transferId})

    @jetapi(min_api=2, auth=True, transferId=str)
    def suspendTransferAsync(self, transferId):
        """Suspends a transfer.

        Suspending a transfer removes it from the server's processing queue,
        and it will stop sending data to its destination. The transfer is
        kept in its current (incomplete) state until it is resumed, or deleted.

        Only active transfers can be suspended. If a transfer is already complete,
        or has failed, this request will fail.

        Args:
            transferId (int, dict): A transfer ID or a :ref:`structure-transfer`.

        Returns:
            dict::ref:`structure-request` containing a ``requestId`` for this request. Use :meth:`~.jetstream.serverinterface.ServerInterface.getRequestStatus` or :meth:`~.jetstream.serverinterface.ServerInterface.getRequestResult` to monitor the status of the request.

            :meth:`~.jetstream.serverinterface.ServerInterface.getRequestResult` will return :ref:`structure-transfer`.

        Raises:
            CommandError: The transfer was not found, or could not be suspended.

        Example:

            >>> request = send.suspendTransferAsync(transfer)
            {'requestId': 7}
            >>> status = send.getRequestStatus(request)
            {'status': 'complete', ...}
            >>> result = send.getRequestResult(request)
            {'transferId': 13}
            >>> transfer = send.getTransfer(result)
            {'status': 'suspended', ...}

        """
        return self._api_interface._send('send.suspendTransferAsync', {'transferId': transferId})

    @jetapi(min_api=8, auth=True, transferId=str, priority=int, priorityLane=str)
    def setTransferPriorityAsync(self, transferId, priority=None, priorityLane=None):
        """Adjusts the priority and priority lane of a transfer.

        Priority is specified as an unsigned integer from 0 to 4294967295. Lower values
        are considered higher priority; i.e. priority 1 takes precedence over priority 2.
        The default is 100. The priority of a transfer may be adjusted at any time.

        Priority lane is the priority queue which will process the transfer.

        Args:
            transferId (int, dict): A transfer ID or a :ref:`structure-transfer`.
            priority (int, optional): The new priority to assign to the transfer. Accepted values are between 0 and 4294967295, inclusive.
            priorityLane (str, optional): The new priority lane to assign to the transfer.

        Returns:
            dict: Returns a :ref:`structure-request` containing a ``requestId`` for this request. Use :meth:`~.jetstream.serverinterface.ServerInterface.getRequestStatus` or :meth:`~.jetstream.serverinterface.ServerInterface.getRequestResult` to monitor the status of the request.

        Result:
            dict: :meth:`~.jetstream.serverinterface.ServerInterface.getRequestResult` yields a dict with the ``transferId``. Use :meth:`getTransfer()` to retrieve the full :ref:`structure-transfer`.

        Raises:
            CommandError: An invalid priority or priority lane was specified.
            CommandError: The transfer was not found.
        """
        return self._api_interface._send('send.setTransferPriorityAsync', {'transferId': transferId, 'priority': priority, 'priorityLane': priorityLane})

    @jetapi(min_api=23, auth=True, transferId=str, sendRateMax=float)
    def setTransferSendRateMaxAsync(self, transferId, sendRateMax):
        """Adjusts the maximum send rate of a transfer.

        The actual transfer rate of a transfer may be less depending on the destination
        settings and network conditions. A rate of -1 turns off rate throttling for this
        transfer.

        Args:
            transferId (int, dict): A transfer ID or a :ref:`structure-transfer`.
            sendRateMax (float): Specifies the maximum transfer rate in kilobits/second.

        Returns:
            dict: Returns a :ref:`structure-request` containing a ``requestId`` for this request. Use :meth:`~.jetstream.serverinterface.ServerInterface.getRequestStatus` or :meth:`~.jetstream.serverinterface.ServerInterface.getRequestResult` to monitor the status of the request.

        Result:
            dict: :meth:`~.jetstream.serverinterface.ServerInterface.getRequestResult` yields a dict with the commited ``sendRateMax`` value.
        
        Raises:
            CommandError: The transfer was not found.

        .. versionadded:: 1.6.1
        """
        return self._api_interface._send('send.setTransferSendRateMaxAsync', {'transferId': transferId, 'sendRateMax': sendRateMax})

    @jetapi(min_api=2, auth=True, transferId=str)
    def deleteTransfer(self, transferId):
        """Deletes a transfer.

        A transfer may not be deleted while it is active. You must first suspend it
        using :meth:`suspendTransferAsync`.

        Args:
            transferId (int, dict): A transfer ID or a :ref:`structure-transfer`.

        Returns:
            None

        Raises:
            CommandError: The transfer was not found.
        """
        return self._api_interface._send('send.deleteTransfer', {'transferId': transferId})

    @jetapi(min_api=15, auth=True, transferId=str, errorMessage=str)
    def failTransfer(self, transferId, errorMessage=None):
        """Causes an active transfer to fail and present the given error message.

        Args:
            transferId (int, dict): A transfer ID or a :ref:`structure-transfer`.
            errorMessage (string, optional): A message describing why the transfer failed.

        Returns:
            None

        Raises:
            CommandError: The transfer was not found.

        .. versionadded:: 1.4.1
        """
        return self._api_interface._send('send.failTransfer', {'transferId': transferId, 'message': errorMessage})

    @jetapi(min_api=2, auth=True, transferId=str)
    def waitForTransfer(self, transferId):
        """Waits for a transfer to finish sending all data.

        After a transfer is created, it may take some time to send its files
        to the destination. :meth:`waitForTransfer` may be used to wait
        until a transfer has finished sending all files.

        Note:
            This method will block until the transfer is complete.

        Args:
            transferId (int, dict): A transfer ID or a :ref:`structure-transfer`.

        Returns:
            dict: :ref:`structure-transfer`. The ``status`` field will indicate that the transfer is ``complete``.

        Raises:
            CommandError: The transfer was not found.
        """
        return self._api_interface._send('send.waitForTransfer', {'transferId': transferId})

    @jetapi(min_api=18, auth=True, destinationId=str, priority=int, transferFlags=int, checkpointFrequencySeconds=int, sendRateMax=float, priorityLane=str, overwriteMode=int, minCipher=int, fileTimeResolutionMicroseconds=longint)
    def updateTransferDefaults(self, destinationId, priority=None, transferFlags=None, checkpointFrequencySeconds=None, sendRateMax=None, priorityLane=None, overwriteMode=None, minCipher=None, fileTimeResolutionMicroseconds=None):
        """Sets the default values for new transfers.

        Set the default values used when creating new transfer with :meth:`createTransfer`. Only parameters
        that not None will be modified. See :meth:`createTransfer` for more details on the parameters.

        Args:
            destinationId (str, dict): A destination ID or a :ref:`structure-destination`.
            priority: (optional) Specifies the initial priority for the new transfer.
            transferFlags: (optional) A flag which controls whether to enable writes and reads on the destination.
            checkpointFrequencySeconds: (optional) How frequently to take a checkpoint. If 0 (zero), checkpoints are disabled.
            sendRateMax: (optional) Maximum transfer rate in kilobits/second.
            priorityLane: (optional) The priority lane of the transfer.
            overwriteMode: (optional) How to handle creating files if a file with the same name already exists.
            minCipher (int): (optional) Minimum desired encryption type.
            fileTimeResolutionMicroseconds (int): (optional) Files with modification times within the given number of microseconds are considered equal.

        Returns:
            None

        Raises:
            CommandError: The destination was not found.

        .. versionchanged:: 1.7.0 Added ``overwriteMode`` parameter.
        .. versionchanged:: 2.2.0 Added ``minCipher`` parameter.
        """
        return self._api_interface._send('send.updateTransferDefaults', {'destinationId': destinationId, 'priority': priority, 'transferFlags': transferFlags, 'checkpointFrequencySeconds': checkpointFrequencySeconds, 'sendRateMax': sendRateMax, 'priorityLane': priorityLane, 'overwriteMode': overwriteMode, 'minCipher': minCipher, 'fileTimeResolutionMicroseconds': fileTimeResolutionMicroseconds})

    # -- -- -- -- -- -- -- -- -- -- -- -- -- -- -- -- -- -- -- -- -- -- -- -- -- -- -- -- -- --
    # User data -- applies to manifests, transfers and shared tokens

    @jetapi(min_api=14, auth=True, objId=str, key=str, value=str)
    def updateObjectUserData(self, objId, key, value=None):
        """Add or change a piece of user data associated with a destination, manifest, transfer or shared token.

        Args:
            id (str, dict): An object ID or a structure (:ref:`structure-destination`, :ref:`structure-manifest`, :ref:`structure-transfer`, :ref:`structure-token`)
            value (string): (optional) The value to associated with the key.

        Returns:
            None

        Raises:
            CommandError: The user data could not be added or changed.

        .. versionadded:: 1.4.1
        """
        return self._api_interface._send('send.updateObjectUserData', {'id': objId, 'key': key, 'value': value})

    @jetapi(min_api=14, auth=True, objId=str, key=str)
    def deleteObjectUserData(self, objId, key):
        """Remove a piece of user data associated with a destination, manifest, transfer or shared token.

        Args:
            id (int, dict): An object ID or a structure (:ref:`structure-destination`, :ref:`structure-manifest`, :ref:`structure-transfer`, :ref:`structure-token`)

        Returns:
            None

        Raises:
            CommandError: The user data could not be deleted.

        .. versionadded:: 1.4.1
        """
        return self._api_interface._send('send.deleteObjectUserData', {'id': objId, 'key': key})

    # -- -- -- -- -- -- -- -- -- -- -- -- -- -- -- -- -- -- -- -- -- -- -- -- -- -- -- -- -- --
    # Connections

    @jetapi(min_api=25, auth=True, superuser=True, connectionId=str)
    def getConnection(self, connectionId):
        """Returns the configuration for a specified connection.

        A connection describes amalgamate configuration information and performance
        statistics shared by destinations which connect to the same server.

        Args:
            connectionId (int, dict): A connection ID or a :ref:`structure-connection`.

        Returns:
            dict: :ref:`structure-connection`

        .. versionadded:: 1.7.0
        """
        return self._api_interface._send('send.getConnection', {'connectionId': connectionId})

    @jetapi(min_api=25, auth=True, superuser=True, limit=int)
    def getConnections(self, limit=0):
        """Returns the current configurations for all known connections.

        A connection describes amalgamate configuration information and performance
        statistics shared by destinations which connect to the same server.

        Args:
            limit (int): (optional) The maximum number of results to retrieve.
                         If not specified, all available connections are returned.

        Returns:
            list: A list of :ref:`connection structures <structure-connection>`, each describing a connection.

        .. versionadded:: 1.7.0
        """
        return self._api_interface._send('send.getConnections', {'limit': limit})

    @jetapi(min_api=25, auth=True, superuser=True, connectionId=str, sendRateMin=float, sendRateMax=float, sendRate=float, rateControlEnabled=bool, rateControlMethod=int)
    def updateConnectionParamsAsync(self, connectionId, sendRateMin=None, sendRateMax=None, sendRate=None, rateControlEnabled=None, rateControlMethod=None):
        """Update connection parameters.

        Args:
            connectionId (str, dict): A connection ID or a structure :ref:`structure-connection`
            sendRateMin (float): The minimum send rate allowed for this destination, in kilobits-per-second. The rate controller will maintain the send rate at or above this value.
            sendRateMax (float): The maximum send rate allowed for this server, in kilobits-per-second. The rate controller will not allow the send rate to exceed this value.
            sendRate (float): The desired send rate, in kilobits-per-second. With rate control enabled, this value will change depending on decisions made by the rate controller.
            rateControlEnabled (bool): ``True`` enables the rate controller, ``False`` disables it.
            rateControlMethod (int):  Sets the current rate control method. **This parameter is reserved for internal and debugging purposes.**

        Returns:
            None

        Raises:
            CommandError: The connection was not found.


        .. versionadded:: 1.7.0
        """
        return self._api_interface._send('send.updateConnectionParamsAsync', {'connectionId': connectionId, 'sendRateMin': sendRateMin, 'sendRateMax': sendRateMax, 'sendRate': sendRate, 'rateControlEnabled': rateControlEnabled, 'rateControlMethod': rateControlMethod})

    # -- -- -- -- -- -- -- -- -- -- -- -- -- -- -- -- -- -- -- -- -- -- -- -- -- -- -- -- -- --
    # Testing

    @jetapi(min_api=6, auth=True, destinationId=str, pingCount=int)
    def schedulePingAsync(self, destinationId, pingCount):
        """Schedules a series of pings to be sent to a destination.

        Pings are scheduled and interleaved into the transmission protocol at a regulated
        frequency. A response is assigned to the Destination as 'lastReceiverDestinationId'.
        Use getDestination to monitor a ping response.

        Args:
            destinationId (str, dict): A destination ID or a :ref:`structure-destination`.
            pingCount: The number of pings to send.

        Returns:
            dict: Returns a :ref:`structure-request` containing a ``requestId`` for this request. Use :meth:`~.jetstream.serverinterface.ServerInterface.getRequestStatus` or :meth:`~.jetstream.serverinterface.ServerInterface.getRequestResult` to monitor the status of the request.

        Raises:
            CommandError: The destination was not found.
        """
        return self._api_interface._send('send.schedulePingAsync', {'destinationId': destinationId, 'pingCount': pingCount})
