#!/usr/bin/env python
# encoding: utf-8
#
# Copyright 2017 Jet Digital Inc. All rights reserved.
#------------------------------------------------------------------------------------------------

import sys
if sys.version_info[0] > 2:
    longint=int
else:
    longint=long # pylint: disable=undefined-variable

from jetstream.apidecorator import jetapi, jetapi_extract_param

#------------------------------------------------------------------------------------------------
class ServerInterface(object):
    # pylint: disable=protected-access

    def __init__(self, api_interface, sync_interface, logging_interface):
        self._api_interface = api_interface
        self.__logging_interface = logging_interface
        self.api_version = None
        self.addWrangler = sync_interface.create_sync_fn(api_interface, self.addWranglerAsync)
        self.ls = sync_interface.create_sync_fn(api_interface, self.lsAsync)
        self.lsList = sync_interface.create_sync_fn(api_interface, self.lsListAsync)
        self.lsRoots = sync_interface.create_sync_fn(api_interface, self.lsRootsAsync)
        self.mkdir = sync_interface.create_sync_fn(api_interface, self.mkdirAsync)
        self.mkFile = sync_interface.create_sync_fn(api_interface, self.mkFileAsync)
        self.rm = sync_interface.create_sync_fn(api_interface, self.rmAsync)
        self.mv = sync_interface.create_sync_fn(api_interface, self.mvAsync)
        self.du = sync_interface.create_sync_fn(api_interface, self.duAsync)
        self.setLicenseKey = sync_interface.create_sync_fn(api_interface, self.setLicenseKeyAsync)
        self.sendDiscoverPacket = sync_interface.create_sync_fn(api_interface, self.sendDiscoverPacketAsync)
        self.recvDiscoverPacket = sync_interface.create_sync_fn(api_interface, self.recvDiscoverPacketAsync)

    # - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - -
    @jetapi(min_api=13, userName=str, password=str, token=str, tokenSession=str)
    def authAsync(self, userName=None, password=None, token=None, tokenSession=None):
        """Authenticates a user.

        

        If the JetStream server is running as root or daemon, you must call either
        :meth:`~.jetstream.serverinterface.ServerInterface.auth` or :meth:`~.jetstream.serverinterface.ServerInterface.authAsync`
        after calling :meth:`~.jetstream.apiinterface.APIInterface.connect`, before any
        other method is called (except :meth:`~.jetstream.serverinterface.ServerInterface.getServerInfo`)

        Args:
            userName (string): Name of the user on the JetStream server. Must be \
                                None if ``token`` is set.
            password (string): Password for the user or token.
            token (string): Shared token. Must be None if ``userName`` is set.
            tokenSession (string): Shared token session. If ``token`` is set, this parameter \
                              specifies the optional session. Sharing a session between two \
                              token authed connections allows them to share data such as transfers. \
                              If ``tokenSession`` is not specified, one will be generated \
                              automatically. The current token session is returned by \
                              :meth:`~.jetstream.serverinterface.ServerInterface.getServerInfo`.

        Returns:
            dict: Returns a :ref:`structure-request` containing a ``requestId`` for this request. Use :meth:`~.jetstream.serverinterface.ServerInterface.getRequestStatus` or :meth:`~.jetstream.serverinterface.ServerInterface.getRequestResult` to monitor the status of the request.

            :meth:`~.jetstream.serverinterface.ServerInterface.getRequestResult` will return ``None`` on success, or raise a ``CommandError``.

        Tip:
            Use ``getpass.getpass()`` to prompt for the ``password`` parameter interactively.
        """
        return self._api_interface._send('authAsync', {'userName':userName, 'password':password, 'token':token, 'tokenSession':tokenSession})

    # - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - -
    @jetapi(min_api=13, userName=str, password=str)
    def auth(self, userName, password):
        """Authenticates a user.

        Helper method for :meth:`~.jetstream.serverinterface.ServerInterface.authAsync` to make the call synchronous.

        Args:
            userName (string): Name of the user on the JetStream server.
            password (string): Password for the user above. Pass in getpass.getpass() \
                               if you wish to input the password interactively.

        Returns:
            None

        Raises:
            CommandError: On auth failure
        """
        self.getRequestResult(self.authAsync(userName=userName, password=password))

    # - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - -
    @jetapi(min_api=13, token=str, password=str, tokenSession=str)
    def authToken(self, token, password=None, tokenSession=None):
        """Authenticates using a shared token.

        Helper method for :meth:`~.jetstream.serverinterface.ServerInterface.authAsync` to make the call synchronous.

        Args:
            token (string): Shared token as generated by :meth:`~.jetstream.sendinterface.createSharedToken`
            password (string): Password for the token above. Pass in getpass.getpass() \
                               if you wish to input the password interactively.
            tokenSession (string): Shared token session. If ``token`` is set, this parameter \
                              specifies the optional session. Sharing a session between two \
                              token authed connections allows them to share data such as transfers. \
                              If ``tokenSession`` is not specified, one will be generated \
                              automatically. The current token session is returned by \
                              :meth:`~.jetstream.serverinterface.ServerInterface.getServerInfo`.

        Returns:
            None

        Raises:
            CommandError: On auth failure
        """
        self.getRequestResult(self.authAsync(token=token, password=password, tokenSession=tokenSession))

    # - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - -
    @jetapi(min_api=15, auth=True, text=str)
    def echo(self, text=None):
        """Sends the (optionally) given text to the server, and returns it from the server.

        Tip:
            This method can be used to check if the API connection is still active.

        Args:
            text (string): Optional text to be echoed by the server.

        Returns:
            None, str: None if not input text given, or the given text.
        """
        return self._api_interface._send('echo', {'text': text})

    # - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - -
    @jetapi(min_api=3)
    def getServerInfo(self):
        """Returns information about the connected server, including its name and API version.

        Tip:
            This method does not require authentication.

        Returns:
            dict: :ref:`structure-server-info`
        """
        return self._api_interface._send('getServerInfo')

    # - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - -
    @jetapi(min_api=13, auth=True, eventName=str, before=longint, after=longint, limit=int)
    def getEventLog(self, eventName=None, before=None, after=None, limit=None):
        """Returns the JetStream event log.

        The log is retrieved chronologically,
        ordered oldest event to newest. Optional parameters may be provided to filter
        which events are retrieved.

        Args:
            eventName (str): (optional) Filters the results to include only this event type. If None, retrieve all events.
            before (int): (optional) Filters the results to include only events before this time. If None, retrieve all entries to the end of the log.
            after (int): (optional) Filters the results to include only events after this time. If None, retrieve all entries from the start of the log.
            limit (int): (optional) The maximum number of log entries to retrieve. If None, retrieve all entries.

        Returns:
            dict: Dictionary with an ``eventLog`` key containing a list of :ref:`event log structures <structure-eventlog>`.

        Example:

            >>> api.server.getEventLog()
            {'eventLog': [{'_event': 'serverStarted', '_timestamp': 1485547699055816},
                          {'_event': 'destinationCreated',
                           '_timestamp': 1485547708422759,
                           'destinationId': '127.0.0.1:8886'}
                         ]
            }
        """
        return self._api_interface._send('getEventLog', {'eventName': eventName, 'before': before, 'after': after, 'limit': limit})

    # - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - -
    @jetapi(min_api=10)
    def getThirdPartyLicenses(self):
        """Returns the licenses for third-party software components used by JetStream.

        Returns:
            dict: Library name, license text.
        """
        return self._api_interface._send('getThirdPartyLicenses')

    # - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - -
    @jetapi(min_api=9, auth=True, path=str, directory=bool, calculateHashType=str)
    def lsAsync(self, path, directory=None, calculateHashType=None):
        """Submits a request to retrieve a list of remote files, as visible to the connected server.

        The result is provided as a dictionary with two keys:

            - ``path``: The path to the directory being listed.
            - ``files``: The stats for the requested files.

        Each item in the "file" list is a dictionary containing the following keys:

            - ``name``: The name of the file. To build the full path, join the parent ``path`` with this name.
            - ``type``: Whether this file is a regular file (``4``), a directory (``2``) or a symlink (``5``).
            - ``size``: The size, in bytes, of a regular file. For a directory or symlink, this will be 0 (zero).
            - ``mtime``: The file's last modification time as a UNIX Time (seconds since Epoch). Use ``time.ctime`` to convert to a human readable form.
            - ``mtimeus``: The file's last modification time in microseconds since Epoch (see :doc:`/api/timestamps`).
            - ``access``: Combined read (``1 << 2``), write (``1 << 1``) and execute (``1 << 0``) bits that describes the current user's access to this file.
            - ``hash``: The hash value of file contents, if a valid ``calculateHashType`` is specified. Otherwise empty.

        Args:
            path (string): The remote path to list on the connected server.
            directory (bool): (optional) If True, the specified path is a directory, and the \
                        result should return information about the directory itself, rather \
                        than the content of the directory.
            calculateHashType (string): (optional) Calculate a hash of each file. This may take some
                        time as the file needs to be read. Currently supports "MD5" and "S3" as the has type.

        Returns:
            dict: Returns a :ref:`structure-request` containing a ``requestId`` for this request. Use :meth:`~.jetstream.serverinterface.ServerInterface.getRequestStatus` or :meth:`~.jetstream.serverinterface.ServerInterface.getRequestResult` to monitor the status of the request.

        .. versionchanged:: 2.3.0 Added ``mtimeus``.
        .. versionchanged:: 2.6.0 Added ``calculateHashType``.
        .. versionchanged:: 2.6.3 Added "MD5" type for ``calculateHashType``.
        """
        return self._api_interface._send('lsAsync', {'path': path, 'directory': directory, 'calculateHashType': calculateHashType})

    # - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - -
    @jetapi(min_api=35, auth=True, statOnly=bool, calculateHashType=str)
    def lsListAsync(self, pathList, statOnly=None, calculateHashType=None):
        """Submits a request to retrieve a list of remote files, as visible to the connected server.

        The result is provided as a list of dictionary with two keys:

            - ``path``: The path to the directory being listed.
            - ``files``: The stats for the requested files.

        The result list is a direct mapping of the input path list.

        Each item in the "file" list is a dictionary containing the following keys:

            - ``name``: The name of the file. To build the full path, join the parent ``path`` with this name.
            - ``type``: Whether this file is a regular file (``4``), a directory (``2``) or a symlink (``5``).
            - ``size``: The size, in bytes, of a regular file. For a directory or symlink, this will be 0 (zero).
            - ``mtime``: The file's last modification time as a UNIX Time (seconds since Epoch). Use ``time.ctime`` to convert to a human readable form.
            - ``mtimeus``: The file's last modification time in microseconds since Epoch (see :doc:`/api/timestamps`).
            - ``access``: Combined read (``1 << 2``), write (``1 << 1``) and execute (``1 << 0``) bits that describes the current user's access to this file.
            - ``hash``: The hash value of file contents, if a valid ``calculateHashType`` is specified. Otherwise empty.

        Args:
            pathList (string list): A list of remote pathes to list on the connected server.
            statOnly (bool): (optional) If True, when the specified path is a directory, and the \
                        result should return information about the directory itself, rather \
                        than the content of the directory.
            calculateHashType (string): (optional) Calculate a hash of each file. This may take some
                        time as the file needs to be read. Currently supports "MD5" and "S3" as the has type.

        Returns:
            dict: Returns a :ref:`structure-request` containing a ``requestId`` for this request. Use :meth:`~.jetstream.serverinterface.ServerInterface.getRequestStatus` or :meth:`~.jetstream.serverinterface.ServerInterface.getRequestResult` to monitor the status of the request.
        """
        return self._api_interface._send('lsListAsync', {'pathList': pathList, 'statOnly': statOnly, 'calculateHashType': calculateHashType})

    # - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - -
    @jetapi(min_api=9, auth=True)
    def lsRootsAsync(self):
        """Submits a request to retrieve a list of the root folders, as visible to the
        connected server.

        The result is provided as a dictionary with two keys:

            - ``path``: Representing the parent of the root folder, this will be an empty string.
            - ``files``: The stats for each root folder.

        Each item in the "file" list is a dictionary containing the following keys:

            - ``name``: The name of the root folder.
            - ``type``: ``10``, combining drive (``8``) and directory (``2``).
            - ``size``: ``0`` (zero).
            - ``mtime``: The folder's last modification time. Use ``time.ctime`` to convert to a human readable form.
            - ``access``: Combined read (``1 << 2``), write (``1 << 1``) and execute (``1 << 0``) bits that describes the current user's access to this root.

        Returns:
            dict: Returns a :ref:`structure-request` containing a ``requestId`` for this request. Use :meth:`~.jetstream.serverinterface.ServerInterface.getRequestStatus` or :meth:`~.jetstream.serverinterface.ServerInterface.getRequestResult` to monitor the status of the request.
        """
        return self._api_interface._send('lsRootsAsync')

    # - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - -
    @jetapi(min_api=13, auth=True, path=str)
    def mkdirAsync(self, path):
        """Submits a request to create a new directory on the connected server, creating all parent directories as needed.

        Tip:
            This will not fail if the requested directory already exists.

        Args:
            path (string): The remote path to create on the connected server.

        Returns:
            dict: Returns a :ref:`structure-request` containing a ``requestId`` for this request. Use :meth:`~.jetstream.serverinterface.ServerInterface.getRequestStatus` or :meth:`~.jetstream.serverinterface.ServerInterface.getRequestResult` to monitor the status of the request.

            :meth:`~.jetstream.serverinterface.ServerInterface.getRequestResult` will return ``None`` on success, or raise a ``CommandError``.

        .. versionadded:: 1.3.0
        .. versionchanged:: 1.8.0 Removed ``parents`` parameter. Parent directories are always created.
        """
        return self._api_interface._send('mkdirAsync', {'path': path})

    # - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - -
    @jetapi(min_api=33, auth=True, path=str, size=int, fillRandom=bool, modifyExisting=bool)
    def mkFileAsync(self, path, size, fillRandom=False, modifyExisting=False):
        """Submits a request to create a new file on the connected server.

        Args:
            path (string): The remote path to create on the connected server.
            size (int): The size of the newly created file.
            fillRandom (bool): Fill the file with random values.
            modifyExisting (bool): If set, replace any existing file with the same name, otherwise return success.

        Returns:
            dict: Returns a :ref:`structure-request` containing a ``requestId`` for this request. Use :meth:`~.jetstream.serverinterface.ServerInterface.getRequestStatus` or :meth:`~.jetstream.serverinterface.ServerInterface.getRequestResult` to monitor the status of the request.

            :meth:`~.jetstream.serverinterface.ServerInterface.getRequestResult` will return ``None`` on success, or raise a ``CommandError``.

        .. versionadded:: 2.5.1
        """
        return self._api_interface._send('mkFileAsync', {'path': path, 'size': size, 'fillRandom': fillRandom, 'modifyExisting': modifyExisting})

    # - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - -
    @jetapi(min_api=19, auth=True, path=str)
    def rmAsync(self, path):
        """Submits a request to remove a path on the connected server. If the path refers to a directory, all files and directories under that path will be removed.

        Args:
            path (string): The remote path to remove on the connected server.

        Returns:
            dict: Returns a :ref:`structure-request` containing a ``requestId`` for this request. Use :meth:`~.jetstream.serverinterface.ServerInterface.getRequestStatus` or :meth:`~.jetstream.serverinterface.ServerInterface.getRequestResult` to monitor the status of the request.

            :meth:`~.jetstream.serverinterface.ServerInterface.getRequestResult` will return ``None`` on success, or raise a ``CommandError``.

        .. versionadded:: 1.5.0
        """
        return self._api_interface._send('rmAsync', {'path': path})

    # - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - -
    @jetapi(min_api=19, auth=True, path=str, newPath=str)
    def mvAsync(self, path, newPath):
        """Submits a request to move or rename a path on the connected server.

        Args:
            path (string): The remote path to create on the connected server.
            newPath (string): The new name for the path.

        Returns:
            dict: Returns a :ref:`structure-request` containing a ``requestId`` for this request. Use :meth:`~.jetstream.serverinterface.ServerInterface.getRequestStatus` or :meth:`~.jetstream.serverinterface.ServerInterface.getRequestResult` to monitor the status of the request.

            :meth:`~.jetstream.serverinterface.ServerInterface.getRequestResult` will return ``None`` on success, or raise a ``CommandError``.

        .. versionadded:: 1.5.0
        """
        return self._api_interface._send('mvAsync', {'path': path, 'newPath': newPath})

    # - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - -
    @jetapi(min_api=10, auth=True, path=str)
    def duAsync(self, path):
        """Submits a request to retrieve the total size, in bytes, for all files in the
        specified remote path, as visible to the connected server.

        The result is provided as a dictionary with a single key:

            - ``du``: The total size, in bytes, collected from the specified path.

        Args:
            path (string): The remote path to enumerate on the connected server.

        Returns:
            dict: Returns a :ref:`structure-request` containing a ``requestId`` for this request. Use :meth:`~.jetstream.serverinterface.ServerInterface.getRequestStatus` or :meth:`~.jetstream.serverinterface.ServerInterface.getRequestResult` to monitor the status of the request.
        """
        return self._api_interface._send('duAsync', {'path': path})

    # - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - -
    @jetapi(min_api=1, auth=True, requestId=longint)
    def getRequestStatus(self, requestId):
        """Returns the status for an asynchronous request.

        API methods whose name ends with ``Async`` represent asynchronous requests. These
        methods are sent a request which must be processed by the destination server,
        and return a ``requestId``.

        :meth:`getRequestStatus` returns the current status for an asynchronous request,
        where the request is identified with its ``requestId``. When an asynchronous request
        is complete, its result is available to :meth:`getRequestResult`.

        Note:
            Allowing the results from asynchronous requests to be retrieved at an arbitrary time
            means that the result must be cached on the server. To avoid resource exhaustion,
            requests are set to expire after a period a time. Once expired, a request is no
            longer available.

        The ``status`` field in the result indicates whether the request was successful.
        It may be one of three values:

            * ``pending``: The request has not yet been processed.
            * ``success``: The request was successful.
            * ``error``: An unrecoverable error has terminated the request.

        You may use the ``pending`` status to implement your own asynchronous poll behavior
        while your script waits for the request to be processed.

        If the status is ``error`` then the ``errorMessage`` field contains a description
        of the error.

        Args:
            requestId (int, dict): A request ID, as returned from any of the ``Async`` methods.
                Also allowed is the full dict from an ``Async`` method. The request ID
                is extracted from the ``requestId`` key.

        Returns:
            dict: :ref:`structure-request`

        Raises:
            CommandError: The request was not found, either expired, or never issued.
        """
        return self._api_interface._send('getRequestStatus', {'requestId': requestId})

    # - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - -
    @jetapi(min_api=1, auth=True, requestId=longint)
    def getRequestResult(self, requestId):
        """Returns the result for an asynchronous request.

        API methods whose name ends with ``Async`` represent asynchronous requests. These
        methods are sent a request which must be processed by the destination server,
        and return a ``requestId``.

        :meth:`getRequestResult` returns the result from an asynchronous request, where the
        request is identified with its ``requestId``.

        If the request is not yet complete, this *blocks* until the result is available.

        If a request fails, :meth:`getRequestResult` will yield ``None``. Use
        :meth:`getRequestStatus` to determine the cause of the failure.

        Note:
            Allowing the results from asynchronous requests to be retrieved at an arbitrary time
            means that the result must be cached on the server. To avoid resource exhaustion,
            requests are set to expire after a period a time. Once expired, a request is no
            longer available.

        Args:
            requestId (int, dict): A request ID or :ref:`structure-request`.

        Returns:
            dict, None: A dictionary describing the result from the request. The structure of the result varies depending on which API method initiated the request.

        Raises:
            CommandError: The request was not found, either expired, or never issued.
            CommandError: If the API method that initiated this request resulted in an error.

        """
        return self._api_interface._send('getRequestResult', {'requestId': requestId})

    # - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - -
    @jetapi(min_api=11, auth=True, discoverId=longint, destinationAddress=str, destinationPort=int)
    def sendDiscoverPacketAsync(self, discoverId, destinationAddress, destinationPort=None):
        """Sends a discovery packet to a waiting server.

        A JetStream server may not know its address and port as visible to another. This
        function sends a special packet to another server to allow that server to discover
        this server's receiving address and port. This function must be used in conjunction
        with :meth:`recvDiscoverPacketAsync`. See that function for a usage example.

        Args:
            discoverId (int, dict): A discover id as returned by :meth:`recvDiscoverPacketAsync`. \
                Also allowed is the full dict from :meth:`recvDiscoverPacketAsync`. The discover ID \
                is extracted from the ``discoverId`` key.
            destinationAddress (str): The host name or IP of the server that called :meth:`~jetstream.serverinterface.ServerInterface.recvDiscoverPacketAsync`.
            destinationPort (int): (optional) The port of the serer that called :meth:`~jetstream.serverinterface.ServerInterface.recvDiscoverPacketAsync`.

        Returns:
            dict: Returns a :ref:`structure-request` containing a ``requestId`` for this request. Use :meth:`~.jetstream.serverinterface.ServerInterface.getRequestStatus` or :meth:`~.jetstream.serverinterface.ServerInterface.getRequestResult` to monitor the status of the request.

        Note:
            A discovery packet may be lost, so it is prudent to send several packets before concluding
            that the connection failed.

        """
        return self._api_interface._send('sendDiscoverPacketAsync', {'discoverId': discoverId, 'destinationAddress': destinationAddress, 'destinationPort': destinationPort})

    # - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - -
    @jetapi(min_api=11, auth=True, timeoutSeconds=float)
    def recvDiscoverPacketAsync(self, timeoutSeconds=None):
        """Creates a request to wait for an incoming receiver discovery packet.

        A JetStream server may not know its address and port as visible to another. This
        function creates a token called a ``discover id`` and then waits for another
        server to send it a special packet that contains this data. Upon receipt, this
        server can derive the address of the server that sent the data. The server whose
        address is being discovered should send the ``discover id`` using :meth:`sendDiscoverPacketAsync`. 

        Args:
            timeoutSeconds (float): (optional) The number of seconds to wait for an incoming discover
                packet. If unused, a default timeout is used.

        Returns:
            dict: Returns a :ref:`structure-request` containing a ``requestId`` for this request. Use :meth:`~.jetstream.serverinterface.ServerInterface.getRequestStatus` or :meth:`~.jetstream.serverinterface.ServerInterface.getRequestResult` to monitor the status of the request. The dict additionally contains a ``discoverId`` key for use with :meth:`sendDiscoverPacketAsync`. Ultimately, the result from the async function will be a :ref:`structure-discover-result`.

            :meth:`~.jetstream.serverinterface.ServerInterface.getRequestResult` will return dictionary :ref:`structure-discover-result`.

        Example:
            .. code-block:: python

                # Server A wants to transfer data to B, but does not know B's public address.
                # We do, however, have API access to both server A and B, and we know A's public address.
                # Server A can "discover" B's address as follows:

                # 1. Ask server A to listen for a discover packet.
                >>> recvRequest = A.server.recvDiscoverPacketAsync()
                >>> discoverId = recvRequest['discoverId']

                # 2. Ask server B to send a discover packet using the return value form recvDiscoverPacketAsync.
                >>> sendRequest = B.server.sendDiscoverPacketAsync(discoverId, A_addr, A_port)
                >>> B.server.getRequestResult(sendRequest)
                {'discoverPacketSent': True}

                # 3. Wait for the discovery packet to be received or a timeout. On success, returns the address of B.
                >>> A.server.getRequestResult(recvRequest)
                {'port': 54946, 'address': '10.96.64.60'}

        """
        return self._api_interface._send('recvDiscoverPacketAsync', {'timeoutSeconds': timeoutSeconds})

    # - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - -
    @jetapi(min_api=1, auth=True, superuser=True, milliseconds=int)
    def sleep(self, milliseconds):
        """Sleep for specified number of milliseconds.
        """
        return self._api_interface._send('sleep', {'milliseconds': milliseconds})

    # - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - -
    @jetapi(min_api=26, auth=True, superuser=True, force=bool)
    def garbageCollect(self, force=False):
        """Initiates an immediate garbage collection on the connected server.

        A periodic garbage collection is automatically performed on the server. This causes
        an immediate garbage collection, with the option to override the allowed grace period.

        A garbage collection can only be initiated by a superuser.

        Args:
            force: (optional) If set, disregards the grace period for items that are ready
                   to be garbage collected and discard immediately. Default is False.

        Raises:
            CommandError: If the user is not granted superuser privileges.

        .. versionadded:: 1.8.0
        """
        return self._api_interface._send('garbageCollect', {'force': force})

    # - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - -
    @jetapi(min_api=10, auth=True, superuser=True)
    def terminate(self) :
        """Terminate the server.

        The user must be a superuser before shutting down the server.
        """
        return self._api_interface._send('terminate')

    # - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - -
    @jetapi(min_api=11, auth=True)
    def superuser(self):
        """Gives a user permission to access and modify all data, and access admin functions.

        Only a user who has successfully called :meth:`~.jetstream.serverinterface.ServerInterface.auth`
        with superuser eligible credentials will be granted superuser privileges. Users are eligible
        for superuser privileges if they share the same credentials as the user who started the server
        or have been listed as eligible during server configuration.

        Superusers are always wranglers (see :meth:`~.jetstream.serverinterface.ServerInterface.wrangler`).

        Returns:
            None

        Raises:
            CommandError: If the user is not granted superuser privileges.
        """
        return self._api_interface._send('superuser')

    # - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - -
    @jetapi(min_api=32, auth=True)
    def wrangler(self):
        """Gives a user permission to access and modify all data.

        Only a user who has successfully called :meth:`~.jetstream.serverinterface.ServerInterface.auth`
        with wrangler eligible credentials will be granted wrangler privileges. Users are eligible
        for wrangler privileges if they have been given wrangler permissions by a superuser using
        the :meth:`~.jetstream.serverinterface.ServerInterface.addWrangler` function.

        Returns:
            None

        Raises:
            CommandError: If the user is not granted wrangler privileges.

        .. versionadded:: 2.5.0
        """
        return self._api_interface._send('wrangler')

    # - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - -
    @jetapi(min_api=32, auth=True, superuser=True, name=str, isGroup=bool)
    def addWranglerAsync(self, name, isGroup=None):
        """Gives a user or a group wrangler permissions.

        Wranglers can only be modified by a superuser.

        Wranglers have the ability to list and modify objects from any user. See
        (see :meth:`~.jetstream.serverinterface.ServerInterface.wrangler`).

        Args:
            name (string): The user name or group name to be added as a wrangler.
            isGroup (bool): The name represents a group instead of a user.

        Returns:
            None

        Raises:
            CommandError: If the user is not granted superuser privileges, or the user or group
            does not exist.

        .. versionadded:: 2.5.0
        """
        return self._api_interface._send('addWranglerAsync', {'name': name, 'isGroup': isGroup})

    # - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - -
    @jetapi(min_api=32, auth=True, superuser=True, name=str, isGroup=bool)
    def getWranglers(self):
        """Gets the list of wrangler user and groups.

        Wranglers can only be accessed by a superuser.

        Returns:
            dict: Returns a dictionary that contains the list of wrangler users and list of wrangler groups.

        Raises:
            CommandError: If the user is not granted superuser privileges.

        .. versionadded:: 2.5.0
        """
        return self._api_interface._send('getWranglers')

    # - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - -
    @jetapi(min_api=32, auth=True, superuser=True, name=str, isGroup=bool)
    def deleteWrangler(self, identifier, isGroup=None):
        """Delete wrangler permissions from a user or a group.

        Wranglers can only be modified by a superuser.

        Note that wranglers are deleted by user id or group id, not by name. Since multiple
        user names can alias the same id, deleting by name would not necessarily prevent
        access. Note also that superusers cannot be deleted.

        Args:
            identifier (string): The user id or group id of the wrangler to be removed.
            isGroup (bool): The name represents a group instead of a user.

        Returns:
            None

        Raises:
            CommandError: If the user is not granted superuser privileges, the user or group id
            is not a wrangler, or the user or group id represents a superuser.

        .. versionadded:: 2.5.0
        """
        return self._api_interface._send('deleteWrangler', {'id': identifier, 'isGroup': isGroup})

    # - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - -
    @jetapi(min_api=35, auth=True, superuser=True, address=str)
    def setExternalAddress(self, address):
        """Sets the external address for the server.

        The external address can be either a full qualified domain name or an IP address.
        This value is also prompted in client as an option for shared link creation.

        Args:
            address (string): A full qualified domain name or an IP address for server external access.

        Returns:
            None

        Raises:
            CommandError: If the user is not granted superuser privileges

        .. versionadded:: 2.6.3
        """
        return self._api_interface._send('setExternalAddress', {'externalAddress': address})

    # - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - -
    @jetapi(min_api=20, auth=True, superuser=True, userName=str)
    def setSandboxMapping(self, userName, sandboxRoots):
        """Sets the sandbox mappings for a user.

        Sandbox mappings can only be modified by a superuser.

        A sandbox mapping is a dictionary that maps sandbox names to sandbox roots for a particular user.
        This sandbox names provide the roots paths available to the user (see :meth:`~.jetstream.serverinterface.ServerInterface.lsRootsAsync`).
        All sandbox root paths are accessed by prefixing paths with the sandbox name followed by a colon.

        Note:
            New sandbox mapping only take effect for new user sessions.

        See :doc:`jetstream-server:sandboxing/index` in JetStream Server Documentation.

        Args:
            userName (string): The user name of the user to sandbox. Note that the user name may be left blank
                to indicates the system sandbox. Any user who does not have a specific sandbox mapping will inherit
                the sandbox mapping from the system sandbox. The system sandbox may contain placeholders such as ``%u``.
            sandboxRoots (dict): A dictionary of sandbox names to sandbox root paths. Sandbox names may be left
                blank to indicate a root of ``/``. If sandbox roots is False, then the user is not sandboxed.

        Returns:
            None

        Raises:
            CommandError: If the user is not granted superuser privileges, a sandbox path is empty, or a sandbox name contains
                illegal characters.

        .. versionadded:: 1.5.0
        """
        return self._api_interface._send('setSandboxMapping', {'userName': userName, 'sandboxRoots': sandboxRoots})

    # - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - -
    @jetapi(min_api=20, auth=True, superuser=True)
    def getSandboxMappings(self):
        """Gets the sandbox mappings for all users.

        Sandbox mappings can only be accessed by a superuser.

        See :doc:`jetstream-server:sandboxing/index` in JetStream Server Documentation.

        Returns:
            dict: Returns a dictionary that maps user names to sandbox mappings, where a sandbox mapping is
            a dictionary that maps sandbox names to sandbox paths.

        Raises:
            CommandError: If the user is not granted superuser privileges.

        .. versionadded:: 1.5.0
        """
        return self._api_interface._send('getSandboxMappings')

    # - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - -
    @jetapi(min_api=20, auth=True, superuser=True, userName=str)
    def deleteSandboxMapping(self, userName):
        """Deletes the sandbox mappings for a user.

        Sandbox mappings can only be modified by a superuser.

        Warning:
            Deleted sandbox mapping only take effect for new user sessions. Existing user sessions will not be affected.

        See :doc:`jetstream-server:sandboxing/index` in JetStream Server Documentation.

        Args:
            userName (string): The user name of the user to sandbox. Note that the user name may be left blank
                to indicates the system sandbox.

        Returns:
            None

        Raises:
            CommandError: If the user is not granted superuser privileges, or no sandbox exists for the user.

        .. versionadded:: 1.5.0
        """
        return self._api_interface._send('deleteSandboxMapping', {'userName': userName})

    # - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - -
    @jetapi(min_api=24, auth=True, superuser=True, incomingLimitActive=bool, incomingRateLimit=float, incomingMaxConnections=int, borrowOutgoing=bool, outgoingLimitActive=bool, outgoingRateLimit=float, borrowIncoming=bool)
    def updateBandwidthLimits(self, incomingLimitActive=None, incomingRateLimit=None, incomingMaxConnections=None, borrowOutgoing=None, outgoingLimitActive=None, outgoingRateLimit=None, borrowIncoming=None):
        """Sets limits on incoming and outgoing bandwidth utilization.

        All parameters are optional. Any parameters which are unspecified or set to None will be unchanged.

        Bandwidth limits can only be modified by a superuser.

        Note:
            Bandwidth limits specified on the server command-line during startup become locked and cannot be changed. Attempting to update locked limits using this method will fail.

        Warning:
            **Compatibility note:** When bandwidth control settings are enabled via the API or command line parameters on a JetStream server, clients prior to version 1.6.0 (API 22) will not be able to communicate with this server.

        Args:
            incomingLimitActive (bool): Apply limits to incoming data.
            incomingRateLimit (float): The maximum rate that data will be received, in kilobits/second.
            incomingMaxConnections (int): The maximum number of incoming connections permitted.
            borrowOutgoing (bool): If there is an outgoing rate limit, allow unused outgoing bandwidth to be added to the incoming rate limit.
            outgoingLimitActive (bool): Apply limits to outgoing data.
            outgoingRateLimit (float): The maximum rate that data will be sent, in kilobits/second.
            borrowIncoming (bool): If there is an incoming rate limit, allow unused incoming bandwidth to be added to the outgoing rate limit.

        Returns:
            None

        Raises:
            CommandError: If the user is not granted superuser privileges, or if bandwidth changes are disallowed.

        .. versionadded:: 1.6.0
        """
        return self._api_interface._send('updateBandwidthLimits', {
                'incomingLimitActive': incomingLimitActive,
                'incomingRateLimit': incomingRateLimit,
                'incomingMaxConnections': incomingMaxConnections,
                'outgoingLimitActive': outgoingLimitActive,
                'outgoingRateLimit': outgoingRateLimit,
                'borrowIncoming': borrowIncoming,
                'borrowOutgoing': borrowOutgoing
        })

    # - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - -
    @jetapi(min_api=26, auth=True, superuser=True, intervalSeconds=int, timeLimitSeconds=int)
    def updateGarbageCollectSettings(self, intervalSeconds=None, timeLimitSeconds=None):
        """Updates garbage collection.

        All parameters are optional. Any parameters which are unspecified or set to None will be unchanged.

        Garbage collection can only be modified by a superuser.

        Args:
            intervalSeconds (int): The time, in seconds, between each garbage collection. Allowed range is from 60 (1 minute) to 2147483647 (about 68 years).
            timeLimitSeconds (int): The grace period, in seconds, to allow for items that are ready to be garbage collected. If 0, all eligible items will be discarded immediately. If non-zero, keep items that have been ready for garbage collection for less time than specified. Allowed range is from 0 to 2147483647 (about 68 years).

        Returns:
            dict: :ref:`structure-garbage-collect`

        Raises:
            CommandError: If the user is not granted superuser privileges.

        .. versionadded:: 1.8.0
        """
        _intervalSeconds = jetapi_extract_param('intervalSeconds', intervalSeconds)
        _timeLimitSeconds = jetapi_extract_param('timeLimitSeconds', timeLimitSeconds)
        return self._api_interface._send('updateGarbageCollectSettings', {'intervalSeconds': _intervalSeconds, 'timeLimitSeconds': _timeLimitSeconds})

    # - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - -
    @jetapi(min_api=28, auth=True, superuser=True, key=str, validate=bool)
    def setLicenseKeyAsync(self, key, validate=None):
        """Submits a request to set the license key for the server.

        The license key can only be modified by a superuser.

        Args:
            key (string): The new license key.
            validate (bool): Only set the license after validating the key with the license server.

        Returns:
            None

        Raises:
            CommandError: If the user is not granted superuser privileges.

        .. versionadded:: 2.1.0
        """
        return self._api_interface._send('setLicenseKeyAsync', {'key': key, 'validate': validate})

    # -- -- -- -- -- -- -- -- -- -- -- -- -- -- -- -- -- -- -- -- -- -- -- -- -- -- -- -- -- --
    # API tokens

    @jetapi(min_api=31, auth=True, path=str, password=str, timeLimitSeconds=int, description=str, permissions=longint)
    def createAPIToken(self, path=None, password=None, timeLimitSeconds=None, description=None, permissions=None):
        """Creates an API token to allow restricted login access.

        Createa an API token with optional password that can be used to log into the server.
        When logged in, the token will impersonate the user that created it, and all file operations
        will occur as that user. Permissions and sandbox path access, however, can be restricted so
        that the token user does not have all the same rights as the original user.

        Common use cases include creating a token with access only to a specific directory (shared folder),
        or creating a token for use in scripting so that the user's credentials remain private. Tokens
        are time limited, and can be revoked at any time, so can be used as a type of simple user
        management. Keep in mind, however, that file permissions are still restricted to the original
        user.

        The returned structure provides a 'token' string which may be be used by clients to log on.

        Args:
            path: (optional) If provided, the token user will only have access to the given path.
                    If not provided, then the token user will have access to all paths/drives/sandboxes
                    as the user.
            password (str): (optional) A password required to auth this token.
            timeLimitSeconds: (optional) Number of seconds from now before this token expires.
            description: (optional) A description of what this token represents.
            permissions: (optional) Permissions of the token. See constants defined in ``jetstream.constants.Permissions``.

        Returns:
            dict: :ref:`structure-token`

        See Also:
            - :doc:`jetstream-server:using_jetstream/api_tokens`

        .. versionadded:: 2.4.0
        """
        return self._api_interface._send('createAPIToken', {'path': path, 'password': password, 'timeLimitSeconds': timeLimitSeconds, 'description': description, 'permissions': permissions})

    @jetapi(min_api=31, auth=True, token=str)
    def getAPIToken(self, token):
        """Returns a structure describing an API token.

        Args:
            token (str, dict): An API Token or :ref:`structure-token`.

        Returns:
            dict: :ref:`structure-token`

        Raises:
            CommandError: The api token was not found.

        See Also:
            - :doc:`jetstream-server:using_jetstream/api_tokens`

        .. versionadded:: 2.4.0
        """
        return self._api_interface._send('getAPIToken', {'token': token})

    @jetapi(min_api=31, auth=True, limit=int)
    def getAPITokens(self, limit=None):
        """Returns information for all known api tokens.

        Args:
            limit (int): (optional) The maximum number of results to retrieve. If not specified, all available api tokens are returned.

        Returns:
            list: A list of :ref:`api token structures <structure-token>`, each describing an api token.

        See Also:
            - :doc:`jetstream-server:using_jetstream/api_tokens`

        .. versionadded:: 2.4.0
        """
        return self._api_interface._send('getAPITokens', {'limit': limit})

    @jetapi(min_api=31, auth=True, token=str)
    def deleteAPIToken(self, token):
        """Deletes the specified api token.

        An api token may be deleted at any time, even if it is currently being used
        as part of a transfer or log in session. The active transfer will continue and will not be terminated.

        Args:
            token (str, dict): An API Token or :ref:`structure-token`.

        Returns:
            None

        Raises:
            CommandError: The api token was not found.

        .. versionadded:: 2.4.0
        """
        return self._api_interface._send('deleteAPIToken', {'token': token})

    @jetapi(min_api=34, auth=True, token=str, timeLimitSeconds=int, description=str)
    def updateAPIToken(self, token, timeLimitSeconds=None, description=None):
        """Updates the specified API token.

        Args:
            token (str, dict): An API Token or :ref:`structure-token`.
            timeLimitSeconds: (optional) Number of seconds from now before this token expires.
            description: (optional) A description of what this token represents.

        Returns:
            None

        Raises:
            CommandError: The API token was not found.

        .. versionadded:: 2.6.0
        """
        return self._api_interface._send('updateAPIToken', {'token': token, 'timeLimitSeconds': timeLimitSeconds, 'description': description})

    # - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - -
    @jetapi(min_api=40)
    def getIdleReceiverPortAsync(self):
        """Returns receiver port asynchronously to be used for initiating transfer.

        Tip:
            This method does not require authentication.

        Returns:
            dict: :ref:``
        """
        return self._api_interface._send('getIdleReceiverPortAsync')
    
    # - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - -
    @jetapi(min_api=40)
    def getIdleReceiverPort(self):
        """Returns receiver port to be used for initiating transfer.

        Tip:
            This method does not require authentication.

        Returns:
            dict: :ref:``
        """
        return self._api_interface._send('getIdleReceiverPort')