#!/usr/bin/env python
# encoding: utf-8
#
# Copyright 2017 Jet Digital Inc. All rights reserved.
#------------------------------------------------------------------------------------------------

import json
import math
import os
import ssl
import socket
import sys

if sys.version_info[0] > 2:
    longint=int
else:
    longint=long # pylint: disable=undefined-variable

from jetstream.constants import NULL, API_VERSION
from jetstream import errors

class SocketInterface(object):

    __HEADER_MAGIC_SIZE = 2
    __HEADER_PACKET_SIZE = 8
    __MAX_PACKET_SIZE = longint(math.pow(16, __HEADER_PACKET_SIZE))

    def __init__(self):
        self.__socket = None
        self.__serverAddress = None
        self.__serverPort = None
        self.__timeoutSeconds = None
        self.__certificatePath = None
        self.__sslCertificateRequirements = ssl.CERT_REQUIRED

    @property
    def serverAddress(self):
        return self.__serverAddress

    @serverAddress.setter
    def serverAddress(self, serverAddress):
        if serverAddress is not NULL:
            self.__serverAddress = serverAddress

    @property
    def serverPort(self):
        return self.__serverPort

    @serverPort.setter
    def serverPort(self, serverPort):
        if serverPort is not NULL:
            self.__serverPort = serverPort

    @property
    def timeoutSeconds(self):
        return self.__timeoutSeconds

    @timeoutSeconds.setter
    def timeoutSeconds(self, timeoutSeconds):
        if timeoutSeconds is not NULL:
            self.__timeoutSeconds = timeoutSeconds

    @property
    def certificatePath(self):
        return self.__certificatePath

    @certificatePath.setter
    def certificatePath(self, certificatePath):
        if certificatePath is NULL:
            return
        try:
            basestring
        except NameError:
            basestring=str # pylint: disable=redefined-builtin
        if not isinstance(certificatePath, basestring):
            raise ValueError('Argument certificatePath has to be string')
        if not os.path.exists(certificatePath):
            raise IOError('certificatePath %s does not exist' % certificatePath)
        self.__certificatePath = certificatePath

    @property
    def serverSocket(self):
        return self.__socket

    @serverSocket.setter
    def serverSocket(self, sock):
        if sock is not NULL:
            self.__socket = sock

    @property
    def sslCertificateRequirements(self):
        return self.__sslCertificateRequirements

    @sslCertificateRequirements.setter
    def sslCertificateRequirements(self, cert_reqs):
        if cert_reqs is not NULL:
            self.__sslCertificateRequirements = cert_reqs

    def connect(self):
        if self.is_connected():
            self.disconnect()

        s = self.__make_socket()

        self.__setup_socket(s)

        self.__socket = self.secure_socket(s)
        try:
            self.__socket.connect((self.serverAddress, self.serverPort))
        except ssl.SSLError as err:
            if hasattr(err, 'reason') and 'ALERT_PROTOCOL_VERSION' in err.reason:
                raise errors.UnsupportedSSLVersionError('The SSL version required by the JetStream server is not supported by this version of Python/OpenSSL. By default, JetStream servers require at least TLS1.2 support.')

            raise


    def listen(self):
        l = self.__prepare_listen_socket()

        l.bind((self.serverAddress, self.serverPort))
        l.listen(128)

        s, addr = l.accept()

        self.__setup_socket(s)

        self.__serverAddress, self.__serverPort = addr

        self.__socket = self.secure_socket(s)

        l.close()

    def get_listen_socket(self):
        return self.__prepare_listen_socket()

    def secure_socket(self, s):
        if not socket.gethostbyname(self.__serverAddress) == '127.0.0.1':
            s = self.__wrap_socket(s)

        return s

    def disconnect(self):
        if self.is_connected():
            s = self.__socket
            self.__socket = None
            try:
                s.shutdown(socket.SHUT_RDWR)
            except socket.error:
                pass
            try:
                s.close()
            except socket.error:
                pass

    def is_connected(self):
        return self.__socket is not None

    def is_secured(self):
        if not self.is_connected():
            return False

        return isinstance(self.__socket, ssl.SSLSocket)

    def get_server_certificate_info(self):
        if not self.is_secured():
            return None
        return self.__socket.getpeercert() or None

    def get_cipher(self):
        if not self.is_secured():
            return None
        return self.__socket.cipher()

    def get_server_certificate(self):
        if not self.is_secured():
            return None
        return ssl.get_server_certificate((self.serverAddress, self.serverPort))

    def send_packet(self, packet):
        self.__send_packet(packet)

        while True:
            response = self.__recv_packet()

            # Special case, where the the packet the string "_heartbeat",
            # this is a "keep alive" response that keeps the socket alive.
            # The real response is still coming.. we return immediately to
            # recv mode.
            if isinstance(response, dict) and ('_heartbeat' in response):
                continue

            break

        try:
            success = response['success']
            message = response['message']
        except (KeyError, TypeError):
            raise errors.ProtocolError('Unexpected response format')

        if not success:
            if 'errorType' in response:
                error_name = response['errorType']
                if not error_name.endswith('Error'):
                    error_name += 'Error'

                cls = None
                try:
                    cls = getattr(errors, error_name)
                except AttributeError:
                    pass

                if cls:
                    raise cls(message)

            raise errors.CommandError(message)

        # Return None if output does not supply the 'result' key.
        return response.get('return')

    # Internal functions
    def __make_socket(self):
        if self.serverAddress is None:
            raise errors.NotConnectedError('Server address has not been specified')
        if self.serverPort is None:
            raise errors.NotConnectedError('Server port has not been specified')

        s = socket.socket(socket.AF_INET, socket.SOCK_STREAM)

        return s

    def __prepare_listen_socket(self):
        l = self.__make_socket()

        if hasattr(socket, 'SO_REUSEADDR'):
            l.setsockopt(socket.SOL_SOCKET, socket.SO_REUSEADDR, 1) # pylint: disable=E1101
        if hasattr(socket, 'SO_REUSEPORT'):
            l.setsockopt(socket.SOL_SOCKET, socket.SO_REUSEPORT, 1) # pylint: disable=E1101
        l.setsockopt(socket.IPPROTO_TCP, socket.TCP_NODELAY, 1) # pylint: disable=E1101

        return l

    def __setup_socket(self, s):
        s.settimeout(self.timeoutSeconds) # pylint: disable=E1101
        s.setsockopt(socket.IPPROTO_TCP, socket.TCP_NODELAY, 1) # pylint: disable=E1101

    def __wrap_socket(self, s):
        params = {
                'cert_reqs': self.sslCertificateRequirements
        }

        if self.certificatePath:
            # If certificate path is specified, we force strict checks
            params['cert_reqs'] = ssl.CERT_REQUIRED
            params['ca_certs'] = self.certificatePath

        # context = ssl.SSLContext(ssl.PROTOCOL_TLSv1_2)
        # context.verify_mode = ssl.CERT_REQUIRED
        # context.check_hostname = True
        # context.load_default_certs()
        
        # ok:ssl-wrap-socket-is-deprecated
        return ssl.wrap_socket(s, **params)

    def __send_packet(self, packet):
        if not self.is_connected():
            raise errors.NotConnectedError()

        data = packet
        p = json.dumps(data)
        if len(p) > SocketInterface.__MAX_PACKET_SIZE:
            raise errors.PacketTooBigError()

        header = 'JS{:>{:>02d}x}'.format(len(p), SocketInterface.__HEADER_PACKET_SIZE)
        try:
            self.__socket.sendall((''.join([header, p]).encode('utf-8')))
        except IOError as err:
            self.disconnect()
            raise err

    def __recv_all(self, desired_size):
        data = []
        size = 0
        while size < desired_size:
            try:
                value = self.__socket.recv(desired_size - len(data))
                if len(value) == 0:
                    raise IOError('Server disconnected')
                data.append(value.decode('utf-8'))
                size += len(value)
            except IOError as err:
                self.disconnect()
                raise err
        return ''.join(data)

    def __recv_packet(self):
        if not self.is_connected():
            raise errors.NotConnectedError()

        # Read the header.
        header = self.__recv_all(SocketInterface.__HEADER_MAGIC_SIZE)
        if header != "JS":
            raise errors.ProtocolError()

        header = self.__recv_all(SocketInterface.__HEADER_PACKET_SIZE)

        try:
            packet_size = longint(header, 16)
        except ValueError:
            raise errors.ProtocolError()

        # Read the data.
        data = self.__recv_all(packet_size)
        try:
            packet = json.loads(data)
        except ValueError as exc:
            # Simulated packet in lieu of expected JSON.
            packet = {'success': False, 'message': 'Sender protocol error. Failed to parse response: {0!s}'.format(exc)}
        return packet

    @staticmethod
    def create_packet(command, parameters):
        packet = {'request': command, 'apiVersion': API_VERSION}
        if parameters:
            try:
                items = parameters.iteritems()
            except AttributeError:
                items = parameters.items()
            packet['parameters'] = dict((key, value) for key, value in items if value is not None)
        return packet
