#!/usr/bin/env python
# encoding: utf-8
# Copyright 2016 Jet Digital Inc. All rights reserved.
#------------------------------------------------------------------------------------------------

class SyncInterface(object):
    """
    SyncInterface provides a procedural way to create a synchronous wrapper for an asynchronous
        API method. It returns a wrapper which performs the asynchronous request, then blocks until
    the request's result is returned.

    The wrapper first calls the asynchronous request function. This returns a RequestId. The
    wrapper then calls getRequestResult(requestId), which blocks until the request is processed
        and the result is available. The result is returned to the client.

    Note:
        Timeout currently unsupported / unused.

        An optional timeout may be specified by the client by adding a 'timeoutSeconds' keyword
        parameter when calling the synchronous function. This parameter is interpreted automatically
        by the SyncInterface; the asynchronous interface, naturally, does not need to implement this.
        If not specified, the SyncInterface blocks until the command is complete.
    """

    def __init__(self):
        self.__timeoutSeconds = None

    # Timeout currently unsupported / unused.
    @property
    def timeoutSeconds(self):
        return self.__timeoutSeconds

    # Timeout currently unsupported / unused.
    @timeoutSeconds.setter
    def timeoutSeconds(self, timeoutSeconds):
        self.__timeoutSeconds = timeoutSeconds

    def create_sync_fn(self, api, async_fn):
        """
        Args:
            api: The API instance.
            async_fn: The asynchronous request function. May have any argument signature.

        Returns:
            dict: Returns a synchronous wrapper for the asynchronous API method.
        """
        return lambda *args, **kwargs : self.sync(api, async_fn, *args, **kwargs)

    @staticmethod
    def sync(api, async_fn, *args, **kwargs):
        """
        Args:
            api: The API instance.
            async_fn: The asynchronous request function.
            *args: The client positional arguments for the asynchronous method.
            *kwargs: The client keyword arguments for the asynchronous method.

        Returns:
            dict: Returns the result from the completed request.
        """
        # Timeout currently unsupported / unused.
        # timeoutSeconds = kwargs.pop('timeoutSeconds', self.timeoutSeconds)
        result = async_fn(*args, **kwargs)
        requestId = result['requestId']
        return api.server.getRequestResult(requestId)

