#!/usr/bin/env python
# encoding: utf-8
#
# Copyright 2017 Jet Digital Inc. All rights reserved.
#------------------------------------------------------------------------------------------------

from copy import copy
import argparse
import getpass
import logging
import os
import os.path
import re
import socket
import struct
import sys
import time

from jetstream import JetStream
from jetstream.constants import DEFAULT_SERVER_ADDRESS, DEFAULT_API_PORT
from jetstream.errors import JetStreamClientException, FileSystemError
from jetstream.tools.progress import NoProgress, ProgressBar, BatchProgress
from jetstream.tools.units import Time, Rate, Size

JETSTREAM_MULTIPLEX_SENDER_PORT = -1

#-----------------------------------------------------------------------------------------------

class InvalidCommandParameters(Exception):
    pass

class JSCPError(Exception):
    pass

class IncompatibleServerVersion(JSCPError):
    pass

#-----------------------------------------------------------------------------------------------

_epilog = '''The JetStream client from which the source file is copied is specified as part of the
source path. For example:

    jscp hostname:/path/to/source ...

The client host may be specified as a name or an IP:

    jscp repo:/path/to/source ...
    jscp 10.10.1.2:/path/to/source ...

If you are copying a file from your local workstation, you can omit the host.
In this case, localhost will be assumed:

    jscp /path/to/source ...

The above is the same as:

    jscp localhost:/path/to/source ...

The JetStream server to which the file is copied is specified as part of the destination
path. For example:

    jscp ... hostname:/path/to/destination

The destination host may be specified as a name or an IP:

    jscp ... jetstream1:/path/to/destination
    jscp ... 10.10.64.8:/path/to/destination

Example use:

    jscp -r 250000 /local/jetstream/jscp.py jetstream1:jetstream/jscp.py

This copies the file "/local/jetstream/jscp.py" from the local JetStream client ("localhost"
is implied here) to the path "jetstream/jscp.py" at the JetStream server named "jetstream1".
The send rate is limited to 250000 kbps.

The progress output shows a progress bar, eta (when available), transfer status, rates,
file size progress, and percent complete. The reported rates are effective rate, average
rate, and maximum allowed rate.

 ''' # This extra space provides a gap between the epilog and the returned prompt.

#-----------------------------------------------------------------------------------------------
class Configuration(object):
    "JSCP configuration structure."
    def __init__(self):
        self.destination_user = None
        self._destination_password = None
        self.destination_address = DEFAULT_SERVER_ADDRESS
        self.destination_api_port = DEFAULT_API_PORT
        self.destination_recv_port = DEFAULT_API_PORT
        self.destination_path = None
        self.send_rate_max = None
        self.source_user = None
        self._source_password = None
        self.source_address = DEFAULT_SERVER_ADDRESS
        self.source_api_port = DEFAULT_API_PORT
        self.source_send_port = None
        self.verbose = False

    @property
    def source_password(self):
        if not self.source_user:
            return None

        if not self._source_password or self._source_password is True:
            print('Enter the password for {:s}@{:s}'.format(self.source_user, self.source_address))
            self._source_password = getpass.getpass('Password: ')

        return self._source_password

    @source_password.setter
    def source_password(self, password):
        self._source_password = password

    @property
    def destination_password(self):
        if not self.destination_user:
            return None

        if not self._destination_password or self._destination_password is True:
            print('Enter the password for {:s}@{:s}'.format(self.destination_user, self.destination_address))
            self._destination_password = getpass.getpass('Password: ')

        return self._destination_password

    @destination_password.setter
    def destination_password(self, password):
        self._destination_password = password

class Stats(object):
    def __init__(self):
        self.effective_rate = Rate()
        self.elapsed = Time()
        self.reported_rate = Rate()
        self.status = None
        self.transfer_size = Size()
        self.queue_size = 0
        self.send_rate_max = Rate()

#-----------------------------------------------------------------------------------------------
class JSCP(object):
    """To use:

        configuration = Configuration()
        # configure to taste.

        jscp = JSCP()
        jscp.set_configuration(configuration)

        # Set desired progress UI. Instance must be derived from Progress.
        jscp.set_progress_ui(...)

        jscp.copy(src_path, dst_path)
    
    """
    def __init__(self):
        # configuration
        self.configuration = Configuration()
        # transfer
        self.destination = None
        self.manifest = None
        self.transfer = None
        self.send_api = None
        self.recv_api = None
        # stats
        self.stats = Stats()
        # UI
        self.progress_ui = NoProgress()

        self.log = logging.getLogger('jetstream.tools.jscp')
        self.log.setLevel(logging.INFO)
        self.log.addHandler(logging.NullHandler())

    # - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - -
    def set_configuration(self, configuration):
        self.configuration = copy(configuration)

        if self.configuration.verbose:
            self.setup_logging()

    # - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - -
    def set_progress_ui(self, progress_ui):
        self.progress_ui = progress_ui
    
    def setup_logging(self):
        log_handler = logging.StreamHandler(sys.stderr)
        log_handler.setLevel(logging.INFO)

        self.log.addHandler(log_handler)

    def setup_api(self):
        # Setup send API
        self.send_api = self.create_api(self.configuration.source_address, self.configuration.source_api_port)

        # Check send API compatibility
        send_api_version = self.send_api.server.getServerInfo()['apiVersion']

        if send_api_version < 22:
            raise IncompatibleServerVersion("This version of JSCP is not able to send files from servers prior to JetStream 1.6.0 (API 22).")

        self.authenticate(self.send_api, self.configuration.source_user, self.configuration.source_password)

        # Setup recv API

        recv_api_address = self.configuration.destination_address

        self.recv_api = JSCP.create_api(recv_api_address, self.configuration.destination_api_port)
        JSCP.authenticate(self.recv_api, self.configuration.destination_user, self.configuration.destination_password)

    # - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - -
    def normalize_paths(self, src_path, dst_path):

        # dst_path must be absolute. This won't strip a trailing slash (for the next test).
        dst_path = os.path.join('/', dst_path)

        # If we don't know if specified destination is file or directory
        if dst_path[-1] != os.sep:
            req = self.recv_api.server.lsAsync(dst_path, directory=True)
            try:
                res = self.recv_api.server.getRequestResult(req)
            except FileSystemError:
                res = {}

            files = res.get('ls', {}).get('files', [])

            # 2 is a magic number for directory
            if len(files) == 1 and files[0]['type'] == 2:
                dst_path += os.sep

        # Do this before dealing with source path, in case it's symlink and realpath() changes the filename
        if dst_path[-1] == os.sep:
            dst_path = os.path.join(dst_path, os.path.basename(src_path))

        # Figure out if the local server is sandboxed. If so, don't give the paths local treatement
        server_info = self.send_api.server.getServerInfo()

        if not server_info['user'].get('sandboxed', False) and is_localhost_ip(self.configuration.source_address):
            src_path = os.path.realpath(src_path)

        return src_path, dst_path

    # - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - -
    def copy(self, src_path, dst_path):
        self.clear()
        try:
            self.setup_api()

            src_path, dst_path = self.normalize_paths(src_path, dst_path)

            self.create_destination()
            self.create_manifest(src_path, dst_path)
            self.create_transfer()
            self.run_transfer()
            self.check_transfer()
            self.show_stats()
        except JetStreamClientException as e:
            raise JSCPError(e)
        except KeyboardInterrupt:
            self.progress_ui.stdout.write('Cancelling...\n')
            raise
        finally:
            self.shutdown()

    # - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - -
    def clear(self):
        # transfer
        self.destination = None
        self.manifest = None
        self.transfer = None
        self.send_api = None
        self.recv_api = None
        # stats
        self.stats = Stats()

    # - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - -
    @staticmethod
    def create_api(address, port):
        jetstream = JetStream()
        api = jetstream.api(address, port)
        try:
            api.connect()
        except socket.error:
            raise JSCPError('JetStream server at {}:{} is not available, or is not allowing connections on the configured port'.format(address, port))
        return api

    # - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - -
    @staticmethod
    def authenticate(api, username, password):
        if not username:
            return

        try:
            api.server.auth(username, password)
        except socket.error:
            raise JSCPError('''Can't authenticate with user '{:s}' with the provided password.'''.format(username))

    # - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - -
    def create_destination(self):

        recv_address = None
        recv_port = None
        sender_port = self.configuration.source_send_port

        do_discovery = self.need_discovery(recv_address)

        discovery = {}

        if do_discovery:
            discovery = self.do_discovery(recv_address)
        else:
            self.log.info('Skipping disocvery of destination server address')

        recv_address = self.configuration.destination_address
        recv_port = self.configuration.destination_recv_port

        if 'result' in discovery and 'address' in discovery['result']:
            recv_address = discovery['result']['address']
            recv_port = discovery['result']['port']
            sender_port = JETSTREAM_MULTIPLEX_SENDER_PORT
            self.log.info('Using discovered address %s:%s', recv_address, recv_port)

        self.destination = self.send_api.send.createDestination(
                recv_address,
                destinationPort=recv_port,
                senderPort=sender_port,
        )

        # reply to dicovery, if necessary to help with opening ports through NAT
        self.discover_reply(discovery)

    def need_discovery(self, recv_address=None):
        src_addr = self.configuration.source_address
        dst_addr = recv_address or self.configuration.destination_address

        # If the addresses are the same, no need to do discovery
        if src_addr == dst_addr:
            return False

        src_is_localhost = is_localhost_ip(src_addr)
        dst_is_localhost = is_localhost_ip(dst_addr)

        # If source IP is localhost, we can't do discovery, as the destination server
        # won't be able to talk back. On machines with multiple IPs, we have no way
        # of identifying which IP will be used to talk to the receiving server
        # (depends on routes). We can't use receiving server to discover it, as that
        # will report the public IP, which may or may not be the same as the IP on
        # the sending server (in case where it's behind firewall, proxy, or NAT). 
        if src_is_localhost:
            return False

        # If both addresses are in 127.0.0.0/8, skip discovery, they're local.
        if src_is_localhost and dst_is_localhost:
            return False

        src_is_private = is_private_ip(src_addr)
        dst_is_private = is_private_ip(dst_addr)

        # If sender is private IP and destination is public, skip discovery, as they
        # won't be able to talk back to us. See the note above attached to local host.
        # Similarly, for sending between two remote servers, we can't know their IP
        # setup.
        if src_is_private and not dst_is_private:
            return False

        # For anything else, do discovery. This includes two private or public IPs.
        # They could be on separate subnets separated by firewall. This could be
        # further expanded to check subnets, but it's simpler to assume discovery is
        # needed. For local network, it's fast and don't introduce much delay. 
        return True


    def do_discovery(self, send_address=None):
        '''
        Perform discovery of receiver's IP from perspective of sender
        '''

        send_address = send_address or self.configuration.source_address
        send_recv_port = self.send_api.server.getServerInfo().get('receiver', {}).get('receiverPort')

        if not send_recv_port:
            raise JSCPError("Could not get server information from source server")

        request = None
        result = None
        success = False

        try:

            # Ask send server to listen for a discover packet.
            self.log.info("Discovering destination server's IP")
            request = self.send_api.server.recvDiscoverPacketAsync()

            while True:
                disc_req = self.recv_api.server.sendDiscoverPacketAsync(request, send_address, send_recv_port)

                self.recv_api.server.getRequestResult(disc_req)
                # Wait for the discovery packet to be receiver or a timeout.

                ready = False
                for _ in range(0,2):
                    s = self.send_api.server.getRequestStatus(request)

                    if s['status'] != "pending":
                        ready = True

                    result = self.send_api.server.getRequestResult(request)
                    time.sleep(1)

                if ready:
                    success = True
                    break

                if not 'discoverPacketSent' in result:
                    success = False
                    raise JSCPError('Discovery of receiver address failed')


        except JSCPError:
            pass

        return {
                'request': request,
                'result': result,
                'success': success,
        }


    def discover_reply(self, discovery=None):
        """
        If there is disocovery information, reply to the discovery packets, in an attempt to open firewall ports.
        """

        if not discovery:
            return

        send_address = self.configuration.source_address
        send_port = self.destination['senderPort']
        self.log.info('Sending discovery reply to %s:%s', send_address, send_port)

        # Punch a hole from the receiver to the sender port.
        result = self.recv_api.server.getRequestResult(
            self.recv_api.server.sendDiscoverPacketAsync(discovery['request'], send_address, send_port)
        )

        if not 'discoverPacketSent' in result:
            #TODO: nice error. include whatever error comes from the request?
            raise JSCPError('Could not send dicovery reply')


    def create_manifest(self, src_path, dst_path):

        fileMappings = None

        if isinstance(src_path, (list, set)) and isinstance(dst_path, (list, set)):
            src_paths = set(src_path)
            dst_paths = set(dst_path)

            if len(src_paths) != len(dst_paths):
                raise JSCPError("Could not create manifest file mapping. Source and destination sets are not the same length")

            fileMappings = dict(zip(src_paths, dst_paths))

        else:
            fileMappings = {src_path: dst_path}

        # Create a manifest, which holds the list of files that need to be
        # transferred and where they should be stored at the receiver side.
        self.manifest = self.send_api.send.createManifest(fileMappings)

        # Creating the manifest is fulfilled immediately. However, the underlying file system
        # collection is an asynchronous operation. Use waitForManifest to block until this is complete.
        self.manifest = self.send_api.send.waitForManifest(self.manifest)
        if self.manifest['status'] == 'error':
            raise JSCPError("Could not create manifest. %s" % self.manifest['errorMessage'])

    # - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - -
    def create_transfer(self):
        self.transfer = self.send_api.send.createTransfer(
                destinationId=self.destination,
                manifestId=self.manifest,
                destinationPath=self.configuration.destination_path,
                sendRateMax=self.configuration.send_rate_max,
                userName=self.configuration.destination_user,
                password=self.configuration.destination_password)
        self.stats.transfer_size.set(self.transfer['totalBlocks'] * self.transfer['blockSize'])

    # - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - -
    def run_transfer(self):
        self.progress_ui.set_total(self.stats.transfer_size)

        with self.progress_ui:
            # Pause the progress so ETA does not measure time before transfer starts sending.
            self.progress_ui.pause(self.transfer['status'])

            while True:
                self.build_stats()
                self.show_progress()

                # The 'processing' key becomes False once the transfer has completed.
                if not (self.transfer['processing'] or self.transfer['status'] == 'suspended'):
                    break

                time.sleep(1)
            
    # - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - -
    def build_stats(self):
        # Use getTransfer to monitor the ongoing transfer.
        self.transfer = self.send_api.send.getTransfer(self.transfer)

        transfers = self.send_api.send.getTransfers(self.transfer['destinationId'])
        transfers = [ t['transferId'] for t in transfers if t['status'] in ('pending', 'sending') ]
        try:
            queue_size = transfers.index(self.transfer['transferId'])
        except ValueError:
            queue_size = 0

        # The destination provides stats about its overall transfer performance.
        self.destination = self.send_api.send.getDestination(self.destination)

        current = self.transfer['blocksAcked'] * self.transfer['blockSize']

        if self.transfer['status'] in ('suspended', 'stalled'):
            self.progress_ui.pause(self.transfer['status'])
        # Wait for send before advancing progress.
        elif current:
            # Advancing progress will unpause and start calculating ETA.
            self.progress_ui.advance_to(current)

        self.stats.elapsed = self.progress_ui.get_elapsed()
        self.stats.effective_rate.set(self.transfer['transferThroughputRate'])
        self.stats.reported_rate.set(self.destination['sendRateAverage'])
        self.stats.status = self.transfer['status']
        self.stats.queue_size = queue_size
        self.stats.send_rate_max.set(self.transfer.get('sendRateMax'))
    
    # - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - -
    def show_progress(self):
        self.progress_ui.show(self.stats)

    # - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - -
    def check_transfer(self):
        if self.transfer['errorMessage']:
            if 'Validation' in self.transfer['errorMessage']:
                raise JSCPError('Transfer failed. A validation error was reported, indicating that a previous or ongoing transfer has already created one of the target files with a size that differs from this transfer. Allowing this transfer would corrupt the previous transfer. Choose a different destination path, then retry this transfer.')
            else: 
                raise JSCPError('Transfer failed: {error}'.format(error=self.transfer['errorMessage']))
        elif not self.transfer['endTime']:	
            raise JSCPError('Transfer has stopped. Cause unknown.')

    # - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - -
    def show_stats(self):
        if self.transfer['endTime']:	
            # Transfer stores times in microseconds. Time uses seconds.
            elapsed = Time(self.transfer['endTime'] - self.transfer['startTime']) / 1000000.0
            effective_rate = Rate(self.stats.transfer_size.kilobits / elapsed)
            self.progress_ui.stdout.write('Copied {size!s} in {elapsed!s}, effective send rate {rate}\n'.format(
                size=self.stats.transfer_size, elapsed=elapsed, rate=effective_rate))

    # - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - -
    def shutdown(self):
        if self.send_api:
            if self.transfer:
                try:
                    # This uses the synchronous suspendTransfer wrapper, which
                    # waits until the suspend operation is complete. 
                    self.send_api.send.suspendTransfer(self.transfer)
                except JetStreamClientException:
                    pass
                try:
                    self.send_api.send.deleteTransfer(self.transfer)
                except JetStreamClientException:
                    pass
                self.transfer = None
            
            if self.manifest:
                try:
                    self.send_api.send.deleteManifest(self.manifest)
                except JetStreamClientException:
                    pass
                self.manifest = None

            self.send_api.disconnect()
            self.send_api = None

        if self.recv_api:
            self.recv_api.disconnect()
            self.recv_api = None

#-----------------------------------------------------------------------------------------------
class Command(object):

    def __init__(self):
        self.configuration = Configuration()
        self.src_path = None
        self.dst_path = None
        self.progress_ui = None

    # - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - -
    def main(self, args):
        if self.parse_args(args):
            jscp = JSCP()
            jscp.set_progress_ui(self.progress_ui)
            jscp.set_configuration(self.configuration)
            jscp.copy(self.src_path, self.dst_path)

    # - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - -
    @staticmethod
    def parse_host_path_spec(host_path_spec):
        match = re.search(r"(((?P<user>.*(?=@))@)?(?P<host>[^:]*)(?<=.):)?(?P<file>.*)", host_path_spec)

        host = match.group('host')
        user = match.group('user')
        filename = match.group('file')

        if not filename:
            filename = None

        if not user:
            user = None

        if not host:
            host = 'localhost'

        return user, host, filename

    # - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - -
    def parse_args(self, args):
        parser = argparse.ArgumentParser(
                prog='jscp',
                description='JetStream copy',
                formatter_class=argparse.RawDescriptionHelpFormatter,
                epilog=_epilog)

        # positional arguments:

        parser.add_argument('source', metavar='source', nargs='?',
                help='Source host and path. Format: "<serverAddress>:<path>"')

        parser.add_argument('destination', metavar='destination', nargs='?',
                help='Destination host and path. Format: "<destinationAddress>:<path>"')

        # optional arguments:

        parser.add_argument('-p', '--serverPort', dest='source_api_port', action='store', type=int, default=DEFAULT_API_PORT,
                help='The API control port of a JetStream Sender server. Default: {default}'.format(default=DEFAULT_API_PORT))

        parser.add_argument('-P', '--destinationPort', dest='destination_recv_port', action='store', type=int, default=DEFAULT_API_PORT,
                help='The UDP port for a JetStream Receiver server that is the destination for a transfer. Default: {default}'.format(default=DEFAULT_API_PORT))

        parser.add_argument('-o', '--destinationAPIPort', dest='destination_api_port', action='store', type=int, default=DEFAULT_API_PORT,
                help='The API port of the destination server. Default: {default}'.format(default=DEFAULT_API_PORT))

        parser.add_argument('-r', '--sendRateMax', dest='send_rate_max', action='store', type=int, default=None,
                help='The maximum send rate allowed for this transfer, in kilobits-per-second. The transfer send rate will not exceed this value.')

        parser.add_argument('-s', '--senderPort', dest='source_send_port', action='store', type=int, default=None,
                help='The port to send from.')

        parser.add_argument('-d', '--destinationPath', dest='destination_path', action='store', default=None,
                help='Specifies a directory that will be created as the parent for all files in the transfer. The manifest paths are respected, and added below this destination path.')

        parser.add_argument('-q', dest='quiet', action='store_true',
                help='Quiet mode: disables the progress meter as well as warning and diagnostic messages.')

        parser.add_argument('-b', dest='batch', action='store_true',
                help='Batch mode: uses line-by-line progress (no curses).')

        parser.add_argument('-v', dest='verbose', action='store_true',
                help='Be more verbose when starting transfers.')


        args = parser.parse_args(args)

        if (args.source is None) or (args.destination is None):
            parser.print_help()
            return False

        source_user, source_address, src_path = self.parse_host_path_spec(args.source)

        if not source_user:
            source_user = getpass.getuser()

        if not src_path:
            raise InvalidCommandParameters('Source path missing')

        destination_user, destination_address, dst_path = self.parse_host_path_spec(args.destination)

        # Always get destination_user. Technically it's not required if server is running with --recv-auth-required=0, but it would still be required to do discovery, which happens most of the time. Rather than be super selective, always get the auth credentials.
        if not destination_user:
            destination_user = getpass.getuser()

        if not dst_path:
            dst_path = os.path.expanduser('~')

        self.configuration.source_user = source_user
        self.configuration.source_address = source_address
        self.configuration.source_api_port = args.source_api_port
        self.configuration.source_send_port = args.source_send_port

        self.configuration.destination_user = destination_user
        self.configuration.destination_address = destination_address
        self.configuration.destination_api_port = args.destination_api_port
        self.configuration.destination_recv_port = args.destination_recv_port
        self.configuration.destination_path = args.destination_path

        self.configuration.send_rate_max = args.send_rate_max

        self.configuration.verbose = args.verbose

        self.src_path = src_path
        self.dst_path = dst_path

        if args.quiet:
            self.progress_ui = NoProgress()     # pylint: disable=E0012,redefined-variable-type
        elif args.batch:
            self.progress_ui = BatchProgress()  # pylint: disable=E0012,redefined-variable-type
        else:
            self.progress_ui = ProgressBar(40)  # pylint: disable=E0012,redefined-variable-type
        
        return True


#-----------------------------------------------------------------------------------------------
def is_private_ip(address):

    if is_localhost_ip(address):
        return True

    address = socket.gethostbyname(address)
    address = struct.unpack('!I', socket.inet_pton(socket.AF_INET, address))[0]
    
    private = (
#		[ 2130706432, 4278190080 ], # 127.0.0.0,   255.0.0.0   http://tools.ietf.org/html/rfc3330
        [ 3232235520, 4294901760 ], # 192.168.0.0, 255.255.0.0 http://tools.ietf.org/html/rfc1918
        [ 2886729728, 4293918720 ], # 172.16.0.0,  255.240.0.0 http://tools.ietf.org/html/rfc1918
        [ 167772160,  4278190080 ], # 10.0.0.0,	   255.0.0.0   http://tools.ietf.org/html/rfc1918
    )

    for net in private:
        if (address & net[1]) == net[0]:
            return True
    
    return False

def is_localhost_ip(address):
    address = socket.gethostbyname(address)
    address = struct.unpack('!I', socket.inet_pton(socket.AF_INET, address))[0]
    
    local = [ 2130706432, 4278190080 ] # 127.0.0.0,   255.0.0.0   http://tools.ietf.org/html/rfc3330

    return (address & local[1]) == local[0]

#-----------------------------------------------------------------------------------------------
def main():
    rc = 0
    command = Command()
    try:
        command.main(sys.argv[1:])
    except InvalidCommandParameters as e:
        sys.stderr.write('{}\n'.format(e))
        rc = 1
    except JSCPError as e:
        sys.stderr.write('{}\n'.format(e))
        rc = 2
    except (OSError, IOError) as e:
        sys.stderr.write('{}\n'.format(e))
        rc = 3
    except KeyboardInterrupt:
        rc = -1
    sys.exit(rc)

#-----------------------------------------------------------------------------------------------
if __name__ == '__main__':
    main()

