#!/usr/bin/env python
# encoding: utf-8
#
# Copyright 2016 Twocubed Technologies Inc. All rights reserved.
#------------------------------------------------------------------------------------------------

import datetime
import os
import sys

from jetstream.tools.units import Time, Size

_ANSI_CLEAR_RIGHT = '\033[0K'
_ANSI_MOVE_UP_N   = '\033[%dA'


#------------------------------------------------------------------------------------------------
class Progress(object):
    """
    Base class which defines the interface used to communicate progress from the Sync class.
    To use:

        progress = Progress()
        sync = Sync(progress=progress)
    
    or:
    
        with Progress() as progress:
            sync = Sync(progress=progress)
    
    Derived class can implement the show() method to display the progress while a Sync is running.
    """

    ETA_MIN_PERCENT =  1	# do not calculate ETA until at least this percentage
    ETA_PERCENT     = 10	# calculate ETA after this percentage
    ETA_SECONDS     = 60	# calculate ETA after this many seconds
    
    # - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - -
    def __init__(self):
        self.__start = None
        self.__current = Size()
        self.__total = Size()
        self.__offset = Size()
        self.__pause = False
        self.__status = None
        self.__average = []

    @property
    def stdout(self):
        return sys.stdout
    
    @property
    def stderr(self):
        return sys.stderr

    @property
    def paused(self):
        return self.__pause
    
    # - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - -
    def get_average(self):
        if self.__average:
            return sum(self.__average) / len(self.__average)
        else:
            return 0
    
    # - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - -
    def get_current(self):
        current = min(self.__current, self.__total)
        return current
    
    # - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - -
    def get_elapsed(self):
        td = datetime.datetime.now() - self.__start
        seconds = ((td.days * 86400) + td.seconds + (td.microseconds / 1000000.0))
        return Time(seconds)

    # - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - -
    def get_eta(self):
        "ETA returned in seconds, as a float. If ETA is not available, returns None."
        eta = Time()
        if not self.paused:
            offset = self.get_offset()
            current = self.get_current() - offset
            total = self.get_total() - offset
            percent = (current * 100.0 / total) if total else 0
            elapsed = self.get_elapsed()
            if (percent > Progress.ETA_MIN_PERCENT) and ((percent > Progress.ETA_PERCENT) or (elapsed > Progress.ETA_SECONDS)):
                # 0.5 is added to discourage a time remaining of "0s" until it's appropriate.
                eta = (elapsed * 100.0 / percent) - elapsed + 0.5
        return eta

    # - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - -
    def get_total(self):
        return self.__total
        
    # - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - -
    def get_offset(self):
        offset = self.__offset if isinstance(self.__offset, int) else 0
        # Offset cannot be greater than total.
        offset = min(offset, self.get_total())
        return offset
        
    # - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - -
    def get_percent(self):
        current = self.get_current()
        total = self.get_total()
        return float(current * 100.0 / total) if total else 0
        
    # - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - -
    def pause(self, status):
        if not self.__pause:
            self.__add_average()
        self.__pause = True
        self.__status = status

    # - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - -
    def resume(self, current):
        self.set_current(current)
        self.__start = datetime.datetime.now()
        self.__pause = False

    # - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - -
    def set_status(self, status):
        self.__status = status
    
    # - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - -
    def show(self, stats):
        pass

    # - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - -
    def advance(self, advance):
        self.__current += advance

    # - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - -
    def advance_to(self, current):
        if self.__pause:
            self.resume(current)
        else:
            self.__current.set(max(current, self.__offset))

    # - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - -
    def set_current(self, current):
        self.__current.set(current)
        self.__offset.set(current)

    def set_total(self, total):
        self.__total.set(total)

    # - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - -
    def start(self):
        if self.__start is None:
            self.__start = datetime.datetime.now()
            # Reset average list.
            self.__average = []
            self.__current.set(self.__offset)
        self.__pause = False

    # - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - -
    def stop(self):
        self.__add_average()
        self.__start = None
        self.__pause = False

    # - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - -
    def __add_average(self):
        """When the progress is paused, we don't want the idle time to contribute to the
        average speed. Each time progress is paused or stopped, the average speed since
        the last start time is added to the 'average' list. The representative average
        is the average of all values in this list."""
        # TODO:
        # While 0 (zero) is perhaps a valid contribution to the average,
        # there's insufficient information here for it to be meaningful.
        # It would need to be weighted against 'elapsed', for example,
        # to reduce undue influence on the representative average.
        progress = self.get_current() - self.get_offset()
        if progress:
            self.__average.append(progress / self.get_elapsed())

    # - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - -
    def __enter__(self):
        self.start()
        return self

    # - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - -
    def __exit__(self, exc_type, exc_value, exc_tb):
        self.stop()
        return False

#------------------------------------------------------------------------------------------------
class NoProgress(Progress):
    def __init__(self):
        super(NoProgress, self).__init__()
        self.__redirect = open(os.devnull, 'w')

    @property
    def stdout(self):
        return self.__redirect
    
    @property
    def stderr(self):
        return self.__redirect
    
    def show(self, stats):
        pass

#------------------------------------------------------------------------------------------------
class ProgressBar(Progress):
    "A progress class which shows a progress bar with an ETA."

    # - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - -
    def __init__(self, width):
        if not isinstance(width, int):
            raise TypeError(type(width))
        if width < 1:
            raise ValueError(width)
        super(ProgressBar, self).__init__()
        self.__width = width
        self.__counter = 0

    # - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - -
    def show(self, stats):
        eta = self.get_eta()
        percent = self.get_percent()
        on = int(self.__width * 0.01 * percent)
        off = self.__width - on
        sys.stdout.write(_ANSI_CLEAR_RIGHT)

        if stats.queue_size > 0:
            dots = (self.__counter % 3) + 1
            plural = 's'

            if stats.queue_size == 1:
                plural = ''

            sys.stdout.write(' Waiting on {queue_size} queued transfer{plural}{dot}\n'.format(
                queue_size=stats.queue_size,
                plural=plural,
                dot='.' * dots))
        elif stats.reported_rate:
            sys.stdout.write(' |{on}{off}| {eta!s}  {status} ({effective}/{reported}/{max} {unit})  {current!s}/{total!s}  {percent:.1f}%\n'.format(
                on='█' * on,
                off=' ' * off,
                eta=eta,
                status=stats.status,
                effective=stats.effective_rate.rate,
                reported=stats.reported_rate.rate,
                max=stats.send_rate_max.rate if stats.send_rate_max else 'unlimited',
                unit=stats.send_rate_max.unit,
                current=self.get_current(),
                total=self.get_total(),
                percent=percent))
        else:
            sys.stdout.write(' |{on}{off}| {eta!s}  {status}  {current!s}/{total!s}  {percent:.1f}%\n'.format(
                on='█' * on,
                off=' ' * off,
                eta=eta,
                status=stats.status,
                current=self.get_current(),
                total=self.get_total(),
                percent=percent))
        sys.stdout.write(_ANSI_MOVE_UP_N % 1)
        sys.stdout.flush()

        self.__counter += 1
    
    # - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - -
    def __exit__(self, exc_type, exc_value, exc_tb):
        sys.stdout.write(_ANSI_CLEAR_RIGHT)
        sys.stdout.write('\n')
        sys.stdout.flush()
        return super(ProgressBar, self).__exit__(exc_type, exc_value, exc_tb)


#------------------------------------------------------------------------------------------------
class BatchProgress(Progress):
    "A progress class for batch mode. Writes TAB-delimited fields."

    # - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - -
    def __init__(self):
        super(BatchProgress, self).__init__()
        self.__show_header = True

    def start(self):
        self.__show_header = True
        super(BatchProgress, self).start()

    # - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - -
    def show(self, stats):
        if self.__show_header:
            self.show_header()
            self.__show_header = False
        eta = self.get_eta()
        sys.stdout.write('{status}\t{effective!s}\t{reported!s}\t{current}\t{total}\t{percent:.1f}%\t{eta}\n'.format(
            eta=eta,
            status=stats.status,
            effective=stats.effective_rate,
            reported=stats.reported_rate,
            current=self.get_current(),
            total=self.get_total(),
            percent=self.get_percent()))
        # Flush is important if user is piping output.
        sys.stdout.flush()

    @staticmethod
    def show_header():
        sys.stdout.write('Status\tEffective Send Rate\tReported Send Rate\tCompleted Size\tTotal Size\tPercent\tETA\n')
