#!/usr/bin/env python
# encoding: utf-8
#
# Copyright 2017 Jet Digital Inc. All rights reserved.
#------------------------------------------------------------------------------------------------

import operator

#-----------------------------------------------------------------------------------------------
class Number(object):

    def __init__(self, value=None):
        self.__value = value
        self.set(value)

    @classmethod
    def create(cls, value):
        return cls(value)

    # - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - -
    
    @property
    def empty(self):
        return self.__value is None

    @property
    def value(self):
        return self.__value or 0

    # - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - -
    def set(self, value):
        try:
            self.__value = value.value
        except AttributeError:
            self.__value = value

    # - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - -

    def __nonzero__(self):
        return bool(self.__value)

    def __bool__(self):
        return bool(self.__value)

    def __int__(self):
        return int(self.value + 0.5)

    def __float__(self):
        return float(self.value)

    def __add__(self, other):
        return self.__op(operator.add, other)

    def __sub__(self, other):
        return self.__op(operator.sub, other)

    def __mul__(self, other):
        return self.__op(operator.mul, other)

    def __rmul__(self, other):
        return self.__rop(operator.mul, other)

    def __div__(self, other):
        return self.__op(operator.truediv, other)

    def __rdiv__(self, other):
        return self.__rop(operator.truediv, other)

    def __truediv__(self, other):
        return self.__op(operator.truediv, other)

    def __rtruediv__(self, other):
        return self.__rop(operator.truediv, other)

    def __floordiv__(self, other):
        return self.__op(operator.floordiv, other)

    def __rfloordiv__(self, other):
        return self.__rop(operator.floordiv, other)

    def __gt__(self, other):
        return self.__op(operator.gt, other)

    def __lt__(self, other):
        return self.__op(operator.lt, other)

    def __op(self, op, other):
        try:
            return self.create(op(self.value, other.value))
        except AttributeError:
            return self.create(op(self.value, other))

    def __rop(self, op, other):
        try:
            return op(other.value, self.value)
        except AttributeError:
            return op(other, self.value)

    def __repr__(self):
        return '{cls}({value})'.format(cls=self.__class__.__name__, value=self.__value)

#-----------------------------------------------------------------------------------------------
class Rate(Number):

    def __init__(self, value=None):
        self.__unit = 'kbps'

        super(Rate, self).__init__(value)

    @property
    def unit(self):
        return self.__unit

    @property
    def rate(self):
        return int(self)

    def __str__(self):
        s = ''
        if not self.empty:
            s = '{} {}'.format(self.rate, self.unit)
        return s

#-----------------------------------------------------------------------------------------------
class Size(Number):

    @property
    def kilobits(self):
        return self.value * 8 / 1000

    #------------------------------------------------------------------------------------------------
    def __str__(self):
        s = ''
        if not self.empty:
            factor = None
            unit = None
            abbrevs = (
                (1<<50, 'PiB'),
                (1<<40, 'TiB'),
                (1<<30, 'GiB'),
                (1<<20, 'MiB'),
                (1<<10, 'KiB'),
                (1, 'B'),
            )

            for factor, unit in abbrevs:
                if self.value >= factor:
                    break

            if factor > 1:
                s = '{value:.1f} {unit}'.format(value=float(self.value)/factor, unit=unit)
            else:
                s = '{value} {unit}'.format(value=self.value, unit=unit)
        return s

#-----------------------------------------------------------------------------------------------
class Time(Number):

    @property
    def seconds(self):
        return self.value

    @property
    def milliseconds(self):
        return self.value * 1000.0

    # - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - -
    def __str__(self):
        def _unit(value, unit, pad):
            return '{value:0{pad}d}{unit}'.format(value=value, unit=unit, pad=(2 if pad else 0))

        s = ''
        if not self.empty:
            days = int(self.value / 86400)
            seconds = self.value - (days * 86400)
            hours = int(seconds / 3600)
            seconds -= (hours * 3600)
            minutes = int(seconds / 60)
            seconds -= (minutes * 60)

            pad = False
            if days:
                s += _unit(days, 'd', pad) + ' '
                pad = True
            if hours or pad:
                s += _unit(hours, 'h', pad) + ' '
                pad = True
            if minutes or pad:
                s += _unit(minutes, 'm', pad) + ' '
                pad = True
            if self.value < 1.0:
                s += '{:.2f}s'.format(seconds)
            else:
                s += _unit(int(seconds + 0.5), 's', pad)
        return s

